import asyncio
import logging
import posixpath
import re
import time
from collections.abc import AsyncGenerator
from json import JSONDecodeError
from typing import Callable, ClassVar, Optional

from pycrashreport.crash_report import CrashReportBase, get_crash_report_from_buf
from xonsh.built_ins import XSH
from xonsh.cli_utils import Annotated, Arg

from pymobiledevice3.exceptions import (
    AfcException,
    AfcFileNotFoundError,
    NotificationTimeoutError,
    SysdiagnoseTimeoutError,
)
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.afc import AfcService, AfcShell, path_completer
from pymobiledevice3.services.notification_proxy import NotificationProxyService
from pymobiledevice3.services.os_trace import OsTraceService

SYSDIAGNOSE_PROCESS_NAMES = ("sysdiagnose", "sysdiagnosed")
SYSDIAGNOSE_DIR = "DiagnosticLogs/sysdiagnose"
SYSDIAGNOSE_IN_PROGRESS_MAX_TTL_SECS = 600

# on iOS17, we need to wait for a moment before trying to fetch the sysdiagnose archive
IOS17_SYSDIAGNOSE_DELAY = 3


class CrashReportsManager:
    COPY_MOBILE_NAME = "com.apple.crashreportcopymobile"
    RSD_COPY_MOBILE_NAME = "com.apple.crashreportcopymobile.shim.remote"

    CRASH_MOVER_NAME = "com.apple.crashreportmover"
    RSD_CRASH_MOVER_NAME = "com.apple.crashreportmover.shim.remote"

    APPSTORED_PATH = "/com.apple.appstored"
    IN_PROGRESS_SYSDIAGNOSE_EXTENSIONS: ClassVar = [".tmp", ".tar.gz"]

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown

        if isinstance(lockdown, LockdownClient):
            self.copy_mobile_service_name = self.COPY_MOBILE_NAME
            self.crash_mover_service_name = self.CRASH_MOVER_NAME
        else:
            self.copy_mobile_service_name = self.RSD_COPY_MOBILE_NAME
            self.crash_mover_service_name = self.RSD_CRASH_MOVER_NAME

        self.afc = AfcService(lockdown, service_name=self.copy_mobile_service_name)

    def __enter__(self):
        raise RuntimeError("Use async context manager: `async with ...`")

    async def __aenter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise RuntimeError("Use async context manager: `async with ...`")

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aclose()

    async def aclose(self) -> None:
        await self.afc.aclose()

    async def clear(self, path: str = "/") -> None:
        """
        Clear crash reports under a path.
        """
        undeleted_items = []
        for filename in await self.ls(path, depth=1):
            undeleted_items.extend(await self.afc.rm(filename, force=True))

        for item in undeleted_items:
            # special case of file that sometimes created automatically right after delete,
            # and then we can't delete the folder because it's not empty
            if item != self.APPSTORED_PATH:
                raise AfcException(
                    f"failed to clear crash reports under {path!r}, undeleted items: {undeleted_items}", None
                )

    async def ls(self, path: str = "/", depth: int = 1) -> list[str]:
        """
        List file and folder in the crash report's directory.
        :param path: Path to list, relative to the crash report's directory.
        :param depth: Listing depth, -1 to list infinite.
        :return: List of files listed.
        """
        result = []
        async for item in self.afc.dirlist(path, depth):
            result.append(item)
        return result[1:]  # skip the root path '/'

    async def parse(self, path: str = "/") -> CrashReportBase:
        """
        Parse a crash report file and return the parsed crash report object.

        :param path: Path to a crash report file.
        :return: Parsed crash report object.
        """
        return get_crash_report_from_buf((await self.afc.get_file_contents(path)).decode(), filename=path)

    async def parse_latest(
        self,
        path: str = "/",
        match: Optional[list[str]] = None,
        match_insensitive: Optional[list[str]] = None,
        count: int = 1,
    ) -> list[CrashReportBase]:
        """
        Parse latest top-level crash report(s) under a path, optionally filtered by basename regex patterns.

        Scans the top level of the given path, filters regular files by basename,
        and returns matches sorted by last modification time (newest first).

        :param path: Path whose top-level entries should be considered. Defaults to '/'
        :param match: Case-sensitive regex patterns over report basename.
        :param match_insensitive: Case-insensitive regex patterns over report basename.
        :param count: Maximum number of latest matching reports to parse.
        :return: Parsed crash report objects ordered from newest to oldest.
                 Result length is between 1 and count.
        :raises ValueError: If count < 1 or if no reports match the filters.

        All provided patterns must match.
        """

        def get_match_arguments_description():
            return ", ".join(
                f"{name}={value!r}"
                for name, value in (("match", match_patterns), ("match_insensitive", match_insensitive_patterns))
                if value
            )

        if count < 1:
            raise ValueError("count must be >= 1")

        match_patterns = match or []
        match_insensitive_patterns = match_insensitive or []

        patterns = [re.compile(pattern) for pattern in match_patterns]
        patterns.extend(re.compile(pattern, re.IGNORECASE) for pattern in match_insensitive_patterns)

        matching_entries_by_mtime = sorted(
            [
                (stat["st_mtime"].timestamp(), entry)
                async for entry in self.afc.dirlist(path, depth=1)
                if entry != path
                and (stat := await self.afc.stat(entry)).get("st_ifmt") == "S_IFREG"
                and (not patterns or all(pattern.search(posixpath.basename(entry)) for pattern in patterns))
            ],
            reverse=True,
        )

        if len(matching_entries_by_mtime) == 0:
            match_arguments_description = get_match_arguments_description()
            raise ValueError(
                f"No reports found ({match_arguments_description})"
                if match_arguments_description
                else "No reports found"
            )

        return [await self.parse(report_path) for _, report_path in matching_entries_by_mtime[:count]]

    async def pull(
        self, out: str, entry: str = "/", erase: bool = False, match: Optional[str] = None, progress_bar: bool = True
    ) -> None:
        """
        Pull crash reports from the device.
        :param out: Directory to pull crash reports to.
        :param entry: File or Folder to pull.
        :param erase: Whether to erase the original file from the CrashReports directory.
        :param match: Regex to match against file and directory names to pull.
        :param progress_bar: Whether to show a progress bar when pulling large files.
        """

        async def _callback(src: str, dst: str) -> None:
            self.logger.info(f"{src} --> {dst}")
            if erase:
                await self.afc.rm_single(src, force=True)

        match = None if match is None else re.compile(match)
        await self.afc.pull(entry, out, match, callback=_callback, progress_bar=progress_bar, ignore_errors=True)

    async def flush(self) -> None:
        """Trigger com.apple.crashreportmover to flush all products into CrashReports directory"""
        ack = b"ping\x00"
        service = await self.lockdown.start_lockdown_service(self.crash_mover_service_name)
        assert ack == await service.recvall(len(ack))

    async def watch(self, name: Optional[str] = None, raw: bool = False) -> AsyncGenerator[str, None]:
        """
        Monitor creation of new crash reports for a given process name.

        Return value can either be the raw crash string, or parsed result containing a more human-friendly
        representation for the crash.
        """
        async for syslog_entry in OsTraceService(lockdown=self.lockdown).syslog():
            if (
                (posixpath.basename(syslog_entry.filename) != "osanalyticshelper")
                or (posixpath.basename(syslog_entry.image_name) != "OSAnalytics")
                or not syslog_entry.message.startswith("Saved type ")
            ):
                # skip non-ips creation syslog lines
                continue

            filename = posixpath.basename(syslog_entry.message.split()[-1])
            self.logger.debug(f"crash report: {filename}")

            if posixpath.splitext(filename)[-1] not in (".ips", ".panic"):
                continue

            while True:
                try:
                    crash_report_raw = (await self.afc.get_file_contents(filename)).decode()
                    crash_report = get_crash_report_from_buf(crash_report_raw, filename=filename)
                    break
                except (AfcFileNotFoundError, JSONDecodeError):
                    # Sometimes we have to wait for the file to be readable
                    pass

            if name is None or crash_report.name == name:
                if raw:
                    yield crash_report_raw
                else:
                    yield crash_report

    async def get_new_sysdiagnose(
        self,
        out: str,
        erase: bool = True,
        *,
        timeout: Optional[float] = None,
        callback: Optional[Callable[[float], None]] = None,
    ) -> None:
        """
        Monitor the creation of a newly created sysdiagnose archive and pull it
        :param out: filename
        :param erase: remove after pulling
        :keyword timeout: Maximum time in seconds to wait for the completion of sysdiagnose archive
            If None (default), waits indefinitely
        :keyword callback: optional callback function (form: func(float)) that accepts the elapsed time so far
        """
        start_time = time.monotonic()
        end_time = None
        if timeout is not None:
            end_time = start_time + timeout
        sysdiagnose_filename = await self._get_new_sysdiagnose_filename(end_time)

        if callback is not None:
            callback(time.monotonic() - start_time)

        self.logger.info("sysdiagnose tarball creation has been started")
        await self._wait_for_sysdiagnose_to_finish(timeout)

        if callback is not None:
            callback(time.monotonic() - start_time)

        await self.pull(out, entry=sysdiagnose_filename, erase=erase)

        if callback is not None:
            callback(time.monotonic() - start_time)

    async def _wait_for_sysdiagnose_to_finish(self, end_time: Optional[float] = None) -> None:
        async with NotificationProxyService(self.lockdown, timeout=end_time) as service:
            stop_notification = "com.apple.sysdiagnose.sysdiagnoseStopped"
            await service.notify_register_dispatch(stop_notification)
            try:
                async for event in service.receive_notification():
                    if event["Name"] != stop_notification:
                        continue
                    self.logger.debug(f"Received {event}")
                    await asyncio.sleep(IOS17_SYSDIAGNOSE_DELAY)
                    break
            except NotificationTimeoutError as e:
                raise SysdiagnoseTimeoutError("Timeout waiting for sysdiagnose completion") from e

    async def _get_new_sysdiagnose_filename(self, end_time: Optional[float] = None) -> str:
        sysdiagnose_filename = None
        excluded_temp_files = []

        while sysdiagnose_filename is None:
            try:
                for filename in await self.afc.listdir(SYSDIAGNOSE_DIR):
                    # search for an IN_PROGRESS archive
                    if filename not in excluded_temp_files and "IN_PROGRESS_" in filename:
                        for ext in self.IN_PROGRESS_SYSDIAGNOSE_EXTENSIONS:
                            if filename.endswith(ext):
                                delta = (
                                    await self.lockdown.get_date()
                                    - (await self.afc.stat(posixpath.join(SYSDIAGNOSE_DIR, filename)))["st_mtime"]
                                )
                                # Ignores IN_PROGRESS sysdiagnose files older than the defined time to live
                                if delta.total_seconds() < SYSDIAGNOSE_IN_PROGRESS_MAX_TTL_SECS:
                                    self.logger.debug(f"Detected in progress sysdiagnose {filename}")
                                    sysdiagnose_filename = filename.rsplit(ext)[0]
                                    sysdiagnose_filename = sysdiagnose_filename.replace("IN_PROGRESS_", "")
                                    sysdiagnose_filename = f"{sysdiagnose_filename}.tar.gz"
                                    return posixpath.join(SYSDIAGNOSE_DIR, sysdiagnose_filename)
                                else:
                                    self.logger.warning(f"Old sysdiagnose temp file ignored {filename}")
                                    excluded_temp_files.append(filename)
            except AfcException:
                pass

            if self._check_timeout(end_time):
                raise SysdiagnoseTimeoutError("Timeout finding in-progress sysdiagnose filename")
            await asyncio.sleep(0.1)

    def _check_timeout(self, end_time: Optional[float] = None) -> bool:
        return end_time is not None and time.monotonic() > end_time


class CrashReportsShell(AfcShell):
    @classmethod
    def create(cls, service_provider: LockdownServiceProvider, **kwargs):
        manager = CrashReportsManager(service_provider)
        XSH.ctx["_manager"] = manager
        super(CrashReportsShell, CrashReportsShell).create(service_provider, service=manager.afc)

    def _setup_shell_commands(self):
        super()._setup_shell_commands()
        self._register_arg_parse_alias("parse", self._do_parse)
        self._register_arg_parse_alias("parse-latest", self._do_parse_latest)
        self._register_arg_parse_alias("clear", self._do_clear)

    def _run_manager(self, awaitable):
        return self.afc._runner.run(awaitable)

    def _do_parse(self, filename: Annotated[str, Arg(completer=path_completer)]) -> None:
        """
        Parse and print a crash report by filename.

        :param filename: Path to the crash report file
        """
        print(self._run_manager(XSH.ctx["_manager"].parse(filename)))

    def _do_parse_latest(
        self,
        path: Annotated[str, Arg(nargs="?", completer=path_completer)] = "/",
        match: Annotated[Optional[list[str]], Arg("--match", "-m", action="append")] = None,
        match_insensitive: Annotated[
            Optional[list[str]],
            Arg("--match-insensitive", "-mi", action="append"),
        ] = None,
        count: Annotated[int, Arg("--count", "-n")] = 1,
    ) -> None:
        """Parse latest top-level crash report(s) under a path, ordered by newest first"""
        latest_reports = self._run_manager(
            XSH.ctx["_manager"].parse_latest(
                path=path,
                match=match or [],
                match_insensitive=match_insensitive or [],
                count=count,
            )
        )
        for report in latest_reports:
            print(report)

    def _do_clear(self, path: Annotated[str, Arg(nargs="?", completer=path_completer)] = "/") -> None:
        """Clear crash reports from the device under a path."""
        self._run_manager(XSH.ctx["_manager"].clear(path))



import asyncio
import logging
from inspect import isawaitable
from typing import Optional, Union

from typing_extensions import Self

from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.service_connection import ServiceConnection


class _LazyServiceConnection:
    def __init__(self, owner: "LockdownService") -> None:
        self._owner = owner

    async def _get_conn(self) -> ServiceConnection:
        await self._owner.connect()
        if self._owner._service is None:
            raise ConnectionError("service is not connected")
        return self._owner._service

    def __getattr__(self, name: str):
        async def _call(*args, **kwargs):
            conn = await self._get_conn()
            attr = getattr(conn, name)
            result = attr(*args, **kwargs)
            if isawaitable(result):
                return await result
            return result

        return _call


class LockdownService:
    def __init__(
        self,
        lockdown: LockdownServiceProvider,
        service_name: str,
        is_developer_service: bool = False,
        service: Optional[ServiceConnection] = None,
        include_escrow_bag: bool = False,
    ) -> None:
        """
        :param lockdown: server provider
        :param service_name: wrapped service name - will attempt
        :param is_developer_service: should DeveloperDiskImage be mounted before
        :param service: an established service connection object. If none, will attempt connecting to service_name
        """
        self._is_developer_service = is_developer_service
        self._include_escrow_bag = include_escrow_bag
        self.service_name: str = service_name
        self.lockdown: LockdownServiceProvider = lockdown
        self._service: Optional[ServiceConnection] = service
        self._service_proxy = _LazyServiceConnection(self)
        self.logger: logging.Logger = logging.getLogger(self.__module__)
        # Shared Future: the first connect() call creates it; concurrent callers join it
        # instead of racing to call start_lockdown_service() multiple times.
        self._connect_future: Optional[asyncio.Future] = None

    @property
    def service(self) -> Union[ServiceConnection, _LazyServiceConnection]:
        if self._service is not None:
            return self._service
        return self._service_proxy

    def __enter__(self) -> Self:
        raise RuntimeError("Use async context manager: `async with ...`")

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise RuntimeError("Use async context manager: `async with ...`")

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self) -> None:
        if self._service is not None:
            await self._service.close()
            self._service = None
        self._connect_future = None

    async def connect(self) -> None:
        if self._service is not None:
            return
        if self._connect_future is not None:
            return await self._connect_future
        fut = asyncio.get_running_loop().create_future()
        self._connect_future = fut
        start_service = (
            self.lockdown.start_lockdown_developer_service
            if self._is_developer_service
            else self.lockdown.start_lockdown_service
        )
        try:
            self._service = await start_service(self.service_name, include_escrow_bag=self._include_escrow_bag)
            fut.set_result(None)
        except BaseException as e:
            self._connect_future = None  # allow retry
            if isinstance(e, asyncio.CancelledError):
                fut.cancel()
            else:
                fut.set_exception(e)
            raise



"""XCUITest execution via the new DTX protocol implementation.

Typical usage::

    async with XCUITestService(lockdown) as svc:
        cfg = await TestConfig.create_for(lockdown, runner_bundle_id="com.example.MyAppUITests.xctrunner", target_bundle_id="com.example.MyApp")
        await svc.run(cfg, timeout=300.0)
"""

from __future__ import annotations

import asyncio
import logging
from contextlib import suppress
from dataclasses import dataclass
from typing import Optional, cast

from packaging.version import Version

from pymobiledevice3.dtx import (
    NSURL,
    NSUUID,
)
from pymobiledevice3.dtx.service import DTXProxyService as _DTXProxyService
from pymobiledevice3.dtx_proxy_service import DtxProxyService
from pymobiledevice3.dtx_service import DtxService
from pymobiledevice3.dtx_service_provider import DtxServiceProvider
from pymobiledevice3.exceptions import AppNotInstalledError, ConnectionTerminatedError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.dvt.instruments.dvt_provider import DvtProvider as InstrumentsDvtProvider
from pymobiledevice3.services.dvt.testmanaged.dtx_services import (  # noqa: F401 — re-exported
    ProcessControlService,
    XCTestCaseResult,
    XCTestDriverInterface,
    XCTestManager_DaemonConnectionInterface,
    XCTestManager_IDEInterface,
    XCUITestListener,
)
from pymobiledevice3.services.dvt.testmanaged.xctest_types import (
    XCTestConfiguration,
)
from pymobiledevice3.services.installation_proxy import InstallationProxyService

logger = logging.getLogger(__name__)
XCODE_VERSION = 36


# ---------------------------------------------------------------------------
# Orchestrator
# ---------------------------------------------------------------------------


class _TestManagerProvider(DtxServiceProvider):
    """DTX transport to ``testmanagerd``.

    Registers the three XCTest service classes so the connection can
    instantiate them when either side opens a channel.
    """

    SERVICE_NAME = "com.apple.testmanagerd.lockdown.secure"
    RSD_SERVICE_NAME = "com.apple.dt.testmanagerd.remote"
    OLD_SERVICE_NAME = "com.apple.testmanagerd.lockdown"


class ProxyIdeToDaemonService(DtxProxyService[XCTestManager_IDEInterface, XCTestManager_DaemonConnectionInterface]):
    pass


class ProxyIdeToDriverService(DtxProxyService[XCTestManager_IDEInterface, XCTestDriverInterface]):
    async def _acquire_channel(self) -> _DTXProxyService:
        # Wait for the runner to open the reverse dtxproxy channel.
        logger.debug("Waiting for XCTestDriverInterface from runner ...")

        try:
            remote_svc = await self.dtx.wait_for_proxied_service(XCTestDriverInterface, remote=True, timeout=30.0)
        except asyncio.TimeoutError:
            raise RuntimeError("Timed out waiting for XCTestDriverInterface — runner did not connect") from None
        else:
            return remote_svc.dtxproxy


class ProcessControlChannel(DtxService[ProcessControlService]):
    """Opens the ProcessControl service channel on the DVT connection."""


class XCUITestService:
    """Orchestrates an XCUITest run using :class:`DTXConnection`.

    Service name selection is handled by :class:`_TestManagerProvider` and
    :class:`_DvtProvider` via :meth:`~DtxServiceProvider.service_name_for`,
    which mirrors the pattern used across the codebase:

    - RSD (iOS 17+ / tunnel) → ``RSD_SERVICE_NAME``
    - iOS ≥ 14.0 over lockdown → ``SERVICE_NAME``
    - iOS < 14.0 → ``OLD_SERVICE_NAME`` with SSL-context strip

    Usage::

        svc = XCUITestService(service_provider)
        await svc.run("com.example.MyAppUITests.xctrunner",
                      env={"AUT_BundleID": "com.example.MyApp"})
    """

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        self.lockdown = lockdown

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    async def run(
        self,
        cfg: TestConfig,
        timeout: Optional[float] = None,
        test_done_event: Optional[asyncio.Event] = None,
        listener: Optional[XCUITestListener] = None,
    ) -> None:
        """Run the XCUITest wait for completion.

        :param cfg: TestConfig with test runner and target app information, as well as test selection.
        :param timeout: Seconds to wait for ``_XCT_didFinishExecutingTestPlan``
            after the test plan starts.  ``TimeoutError`` is raised on expiry.
        :param listener: Optional :class:`XCUITestListener` instance to receive
            lifecycle events (test suite/case start/finish, failures, etc.).
        """
        test_done = test_done_event or asyncio.Event()

        sid = NSUUID.uuid4()
        xctest_config = cfg.to_xctestconfiguration(sid, self.lockdown)
        product_major_version = Version(self.lockdown.product_version).major
        bundle_id = cfg.runner_app_info["CFBundleIdentifier"]
        env = cfg.runner_app_env
        args = cfg.runner_app_args
        xctest_path = f"/tmp/{str(sid).upper()}.xctestconfiguration"

        dvt_provider = InstrumentsDvtProvider(self.lockdown)
        control_test_manager_provider = _TestManagerProvider(self.lockdown)
        main_test_manager_provider = _TestManagerProvider(self.lockdown)

        async with dvt_provider, control_test_manager_provider, main_test_manager_provider:
            # Inject per-run context into the TM connections after connect().
            for tm_prov in (control_test_manager_provider, main_test_manager_provider):
                tm_prov.dtx.ctx["xctest_config"] = xctest_config
                tm_prov.dtx.ctx["test_done_event"] = test_done
                if listener is not None:
                    tm_prov.dtx.ctx["xcuitest_listener"] = listener

            # Open control and main dtxproxy channels.
            ctrl_proxy = ProxyIdeToDaemonService(control_test_manager_provider)
            main_proxy = ProxyIdeToDaemonService(main_test_manager_provider)
            process_control_channel = ProcessControlChannel(dvt_provider)
            driver_ch = ProxyIdeToDriverService(main_test_manager_provider)

            async with ctrl_proxy, main_proxy, process_control_channel:
                ctrl_daemon = cast(
                    XCTestManager_DaemonConnectionInterface,
                    ctrl_proxy.service.remote_service,
                )
                main_daemon = cast(
                    XCTestManager_DaemonConnectionInterface,
                    main_proxy.service.remote_service,
                )

                logger.debug("Initializing ctrl session ...")
                ctrl_result = await ctrl_daemon.init_ctrl_session(product_major_version)
                logger.debug("ctrl session result: %r", ctrl_result)

                logger.debug("Initializing main session (session-id=%s) ...", sid)
                main_result = await main_daemon.init_session(product_major_version, sid, xctest_config)
                logger.debug("main session result: %r", main_result)

                # Launch the test runner process.
                launch_args, launch_env, launch_options = _generate_launch_args(
                    product_major_version, sid, cfg.runner_app_info, xctest_path, env, args
                )
                pid = await process_control_channel.service.launch_suspended_process(
                    "", bundle_id, launch_env, launch_args, launch_options
                )
                logger.debug("Launched test runner pid=%d", pid)

                if product_major_version < 17:
                    await asyncio.sleep(1)

                logger.debug("Authorizing test session for pid=%d ...", pid)
                auth = await ctrl_daemon.authorize_test(product_major_version, pid)
                logger.debug("authorize_test result: %r", auth)

                # Wait for the runner to open the reverse dtxproxy channel.
                logger.debug("Waiting for XCTestDriverInterface from runner ...")

                async with driver_ch:
                    driver_iface = driver_ch.remote_service
                    logger.debug("Starting test plan execution ...")
                    await driver_iface.start_executing_test_plan(XCODE_VERSION)

                    timeout_str = f"{timeout:.1f}s" if timeout is not None else "unlimited time"
                    logger.debug("Waiting for test completion (timeout=%s) ...", timeout_str)

                    # Race test-done event against the runner connection dropping.
                    # When the test runner terminates itself (e.g. because a test
                    # case calls Terminate on the xctrunner process), the DTX TCP
                    # connection is reset before _XCT_didFinishExecutingTestPlan
                    # can be sent.  In that case _disconnected fires first and we
                    # raise ConnectionTerminatedError rather than hanging.
                    done_fut = asyncio.ensure_future(test_done.wait())
                    disc_fut = asyncio.ensure_future(main_test_manager_provider.dtx.wait_disconnected())
                    try:
                        done_set, _ = await asyncio.wait(
                            {done_fut, disc_fut},
                            timeout=timeout,
                            return_when=asyncio.FIRST_COMPLETED,
                        )
                    finally:
                        done_fut.cancel()
                        disc_fut.cancel()
                        with suppress(asyncio.CancelledError, Exception):
                            await asyncio.gather(done_fut, disc_fut, return_exceptions=True)

                    if not done_set:
                        raise TimeoutError(f"Test did not finish within {timeout_str}")
                    if not test_done.is_set():
                        raise ConnectionTerminatedError(
                            "Runner DTX connection closed before _XCT_didFinishExecutingTestPlan"
                            " — the test runner likely terminated itself mid-plan"
                        )


@dataclass
class TestConfig:
    runner_app_info: dict

    runner_app_env: Optional[dict] = None
    runner_app_args: Optional[list] = None

    target_app_info: Optional[dict] = None
    target_app_env: Optional[dict] = None
    target_app_args: Optional[list] = None
    tests_to_run: Optional[list] = None
    tests_to_skip: Optional[list] = None

    @staticmethod
    async def create_for(
        service_provider: LockdownServiceProvider, runner_bundle_id: str, target_bundle_id: Optional[str] = None
    ) -> TestConfig:
        """Helper to create a TestConfig with the required runner_app_info and optional target_app_info."""
        runner_app_info: dict
        target_app_info: Optional[dict] = None

        async with InstallationProxyService(lockdown=service_provider) as install_service:
            apps = await install_service.get_apps(
                bundle_identifiers=[runner_bundle_id] + ([target_bundle_id] if target_bundle_id else [])
            )
            if runner_bundle_id not in apps:
                raise AppNotInstalledError(f"No app with bundle id {runner_bundle_id} found")
            runner_app_info = apps[runner_bundle_id]
            if target_bundle_id:
                if target_bundle_id not in apps:
                    raise AppNotInstalledError(f"No app with bundle id {target_bundle_id} found")
                target_app_info = apps[target_bundle_id]

        return TestConfig(runner_app_info=runner_app_info, target_app_info=target_app_info)

    def to_xctestconfiguration(
        self, session_identifier: NSUUID, service_provider: LockdownServiceProvider
    ) -> XCTestConfiguration:
        assert self.runner_app_info, "runner_app_info must be set"

        cfg = {}
        exec_name: str = self.runner_app_info["CFBundleExecutable"]
        assert exec_name.endswith("-Runner"), f"Invalid CFBundleExecutable: {exec_name}"
        config_name = exec_name[: -len("-Runner")]

        cfg["testBundleURL"] = NSURL(None, f"file://{self.runner_app_info['Path']}/PlugIns/{config_name}.xctest")

        cfg["automationFrameworkPath"] = "/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework"
        if Version(service_provider.product_version).major >= 17:
            cfg["automationFrameworkPath"] = (
                "/System/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework"
            )

        if self.target_app_info is not None:
            assert self.target_app_info != self.runner_app_info, (
                "target_app_info must be different from runner_app_info"
            )
            cfg["productModuleName"] = config_name
            cfg["targetApplicationBundleID"] = self.target_app_info.get("CFBundleIdentifier")
            cfg["targetApplicationPath"] = self.target_app_info.get("Path")
            cfg["targetApplicationEnvironment"] = self.target_app_env or {}
            cfg["targetApplicationArguments"] = self.target_app_args or []
            assert cfg["targetApplicationPath"], "targetApplicationPath must be set if target_app_info is provided"
            assert cfg["targetApplicationBundleID"], (
                "targetApplicationBundleID must be set if target_app_info is provided"
            )

        return XCTestConfiguration({
            "testBundleURL": NSURL(None, f"file://{self.runner_app_info['Path']}/PlugIns/{config_name}.xctest"),
            "sessionIdentifier": session_identifier,
            "testsToRun": self.tests_to_run or set(),
            "testsToSkip": self.tests_to_skip or set(),
            "testsMustRunOnMainThread": True,
            "reportResultsToIDE": True,
            "reportActivities": True,
            "testApplicationDependencies": None,
            **cfg,
        })


def _generate_launch_args(
    product_major_version: int,
    test_session_identifier: NSUUID,
    app_info: dict,
    xctest_path: str,
    test_runner_env: Optional[dict] = None,
    test_runner_args: Optional[list] = None,
) -> tuple[list, dict, dict]:
    """Return *(app_args, app_env, app_options)* for launching the test runner.

    Extracted from the old ``launch_test_app`` method so that it can be
    shared between :class:`XCUITestService` and lower-level probe tests.
    """
    app_container = app_info["Container"]
    app_path = app_info["Path"]
    exec_name = app_info["CFBundleExecutable"]
    assert exec_name.endswith("-Runner"), f"Invalid CFBundleExecutable: {exec_name}"
    target_name = exec_name[: -len("-Runner")]

    app_env = {
        "CA_ASSERT_MAIN_THREAD_TRANSACTIONS": "0",
        "CA_DEBUG_TRANSACTIONS": "0",
        "DYLD_FRAMEWORK_PATH": app_path + "/Frameworks:",
        "DYLD_LIBRARY_PATH": app_path + "/Frameworks",
        "MTC_CRASH_ON_REPORT": "1",
        "NSUnbufferedIO": "YES",
        "SQLITE_ENABLE_THREAD_ASSERTIONS": "1",
        "WDA_PRODUCT_BUNDLE_IDENTIFIER": "",
        "XCTestBundlePath": f"{app_path}/PlugIns/{target_name}.xctest",
        "XCTestConfigurationFilePath": app_container + xctest_path,
        "XCODE_DBG_XPC_EXCLUSIONS": "com.apple.dt.xctestSymbolicator",
        "XCTestSessionIdentifier": str(test_session_identifier).upper(),
    }

    if product_major_version >= 11:
        app_env["DYLD_INSERT_LIBRARIES"] = "/Developer/usr/lib/libMainThreadChecker.dylib"
        app_env["OS_ACTIVITY_DT_MODE"] = "YES"
    if product_major_version >= 17:
        app_env["DYLD_FRAMEWORK_PATH"] = f"${app_env['DYLD_FRAMEWORK_PATH']}/System/Developer/Library/Frameworks:"
        app_env["DYLD_LIBRARY_PATH"] = f"${app_env['DYLD_LIBRARY_PATH']}:/System/Developer/usr/lib"
        app_env["XCTestConfigurationFilePath"] = ""  # sent as return value of _XCT_testRunnerReadyWithCapabilities
        app_env["XCTestManagerVariant"] = "DDI"

    if test_runner_env:
        app_env.update(test_runner_env)

    app_args = [
        "-NSTreatUnknownArgumentsAsOpen",
        "NO",
        "-ApplePersistenceIgnoreState",
        "YES",
    ]
    app_args.extend(test_runner_args or [])

    app_options: dict = {"StartSuspendedKey": False}
    if product_major_version >= 12:
        app_options["ActivateSuspended"] = True

    return app_args, app_env, app_options



from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.lockdown_service import LockdownService

SRCFILES = """Baseband
CrashReporter
MobileAsset
VARFS
HFSMeta
Lockdown
MobileBackup
MobileDelete
MobileInstallation
MobileNotes
Network
UserDatabases
WiFi
WirelessAutomation
NANDDebugInfo
SystemConfiguration
Ubiquity
tmp
WirelessAutomation"""


class FileRelayService(LockdownService):
    SERVICE_NAME = "com.apple.mobile.file_relay"

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)
        self.packet_num = 0

    async def stop_session(self):
        self.logger.info("Disconecting...")
        await self.service.close()

    async def request_sources(self, sources=None):
        if sources is None:
            sources = ["UserDatabases"]
        await self.service.send_plist({"Sources": sources})
        while 1:
            res = await self.service.recv_plist()
            if res:
                s = res.get("Status")
                if s == "Acknowledged":
                    z = b""
                    while True:
                        x = await self.service.recv_any()
                        if not x:
                            break
                        z += x
                    return z
                else:
                    print(res.get("Error"))
                    break
        return None



import asyncio
import json
import logging
from collections import UserDict
from typing import Optional

from pymobiledevice3.exceptions import InspectorEvaluateError
from pymobiledevice3.services.web_protocol.session_protocol import SessionProtocol

logger = logging.getLogger(__name__)
console_logger = logging.getLogger("webinspector.console")
heap_logger = logging.getLogger("webinspector.heap")

webinspector_logger_handlers = {
    "log": console_logger.info,
    "info": console_logger.info,
    "error": console_logger.error,
    "debug": console_logger.debug,
    "warning": console_logger.warning,
}


class JSObjectPreview(UserDict):
    def __init__(self, properties: list[dict]):
        super().__init__()
        for p in properties:
            name = p["name"]
            value = p["value"]
            self.data[name] = value


class JSObjectProperties(UserDict):
    def __init__(self, properties: list[dict]):
        super().__init__()
        for p in properties:
            name = p["name"]
            if name == "__proto__":
                self.class_name = p["value"]["className"]
                continue
            # test if a getter/setter first
            value = p.get("get", p.get("set", p.get("value")))
            if value is None:
                continue
            preview = value.get("preview")
            if preview is not None:
                value = JSObjectPreview(preview["properties"])
            elif value.get("className") == "Function":
                value = value["description"]
            else:
                value = value.get("value")
            self.data[name] = value


class InspectorSession:
    def __init__(self, protocol: SessionProtocol, target_id: Optional[str] = None):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        """
        self.protocol = protocol
        self.target_id = target_id
        self.message_id = 1
        self._last_console_message = {}
        self._dispatch_message_responses = {}

        self.response_methods = {
            "Target.targetCreated": self._target_created,
            "Target.targetDestroyed": self._target_destroyed,
            "Target.dispatchMessageFromTarget": self._target_dispatch_message_from_target,
            "Target.didCommitProvisionalTarget": self._target_did_commit_provisional_target,
            "Console.messageAdded": self._console_message_added,
            "Console.messagesCleared": lambda _: _,
            "Console.messageRepeatCountUpdated": self._console_message_repeated_count_updated,
            "Heap.garbageCollected": self._heap_garbage_collected,
        }

        self._receive_task = asyncio.create_task(self._receive_loop())

    @classmethod
    async def create(cls, protocol: SessionProtocol, wait_target: bool = True):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        :param bool wait_target: Wait for target. If not, all operations won't have a window context to operate in
        """
        await protocol.inspector.setup_inspector_socket(protocol.id_, protocol.app.id_, protocol.page.id_)
        target_id = None
        if wait_target:
            while not protocol.inspector.wir_events:
                await asyncio.sleep(0)
            created = protocol.inspector.wir_events.pop(0)
            while "targetInfo" not in created["params"]:
                created = protocol.inspector.wir_events.pop(0)
            target_id = created["params"]["targetInfo"]["targetId"]
            logger.info(f"Created: {target_id}")
        target = cls(protocol, target_id)
        return target

    def set_target_id(self, target_id):
        self.target_id = target_id
        logger.info(f"Changed to: {target_id}")

    async def heap_gc(self):
        return await self.send_command("Heap.gc")

    async def heap_snapshot(self):
        snapshot = await self.send_command("Heap.snapshot")
        if self.target_id is not None:
            snapshot = json.loads(snapshot["params"]["message"])
        snapshot = json.loads(snapshot)["result"]["snapshotData"]
        return snapshot

    async def heap_enable(self):
        return await self.send_command("Heap.enable")

    async def console_enable(self):
        return await self.send_command("Console.enable")

    async def runtime_enable(self):
        return await self.send_command("Runtime.enable")

    async def send_command(self, method: str, **kwargs):
        if self.target_id is None:
            return await self.protocol.send_receive(method, **kwargs)
        else:
            return await self.send_and_receive({"method": method, "params": kwargs})

    async def runtime_evaluate(self, exp: str, return_by_value: bool = False):
        # if the expression is dict, it's needed to be in ()
        exp = exp.strip()
        if exp and exp[0] == "{" and exp[-1] == "}":
            exp = f"({exp})"

        response = await self.send_and_receive({
            "method": "Runtime.evaluate",
            "params": {
                "expression": exp,
                "objectGroup": "console",
                "includeCommandLineAPI": True,
                "doNotPauseOnExceptionsAndMuteConsole": False,
                "silent": False,
                "returnByValue": return_by_value,
                "generatePreview": True,
                "userGesture": True,
                "awaitPromise": False,
                "replMode": True,
                "allowUnsafeEvalBlockedByCSP": False,
                "uniqueContextId": "0.1",
            },
        })

        return await self._parse_runtime_evaluate(response)

    async def navigate_to_url(self, url: str):
        return await self.runtime_evaluate(exp=f'window.location = "{url}"')

    async def send_and_receive(self, message: dict) -> dict:
        if self.target_id is None:
            message_id = await self.protocol.send_command(message["method"], **message.get("params", {}))
            return await self.protocol.wait_for_message(message_id)
        else:
            message_id = await self.send_message_to_target(message)
            return await self.receive_response_by_id(message_id)

    async def send_message_to_target(self, message: dict) -> int:
        message["id"] = self.message_id
        self.message_id += 1
        await self.protocol.send_command(
            "Target.sendMessageToTarget", targetId=self.target_id, message=json.dumps(message)
        )
        return message["id"]

    async def _receive_loop(self):
        while True:
            while not self.protocol.inspector.wir_events:
                await asyncio.sleep(0)

            response = self.protocol.inspector.wir_events.pop(0)
            response_method = response["method"]
            if response_method in self.response_methods:
                self.response_methods[response_method](response)
            else:
                logger.error(f"Unknown response: {response}")

    async def receive_response_by_id(self, message_id: int) -> dict:
        while True:
            if message_id in self._dispatch_message_responses:
                return self._dispatch_message_responses.pop(message_id)
            await asyncio.sleep(0)

    async def get_properties(self, object_id: str) -> JSObjectProperties:
        message = await self.send_command(
            "Runtime.getProperties", objectId=object_id, ownProperties=True, generatePreview=True
        )
        if self.target_id is not None:
            message = json.loads(message["params"]["message"])["result"]
        return JSObjectProperties(message["properties"])

    async def _parse_runtime_evaluate(self, response: dict):
        message = response if self.target_id is None else json.loads(response["params"]["message"])
        result = message["result"]["result"]
        if result.get("subtype", "") == "error":
            properties = await self.get_properties(result["objectId"])
            raise InspectorEvaluateError(
                properties.class_name,
                properties["message"],
                properties.get("line"),
                properties.get("column"),
                properties.get("stack", "").split("\n"),
            )
        elif result["type"] == "bigint":
            return result["description"]
        elif result["type"] == "undefined":
            pass
        elif result["type"] == "object":
            value = result.get("value")
            if value is not None:
                return value

            # TODO: JSObjectProperties()
            preview = result["preview"]
            preview_buf = "{\n"
            for p in result["preview"]["properties"]:
                value = p.get("value", "NOT_SUPPORTED_FOR_PREVIEW")
                preview_buf += f"\t{p['name']}: {value}, // {p['type']}\n"
            if preview.get("overflow"):
                preview_buf += "\t// ...\n"
            preview_buf += "}"
            return f"[object {result['className']}]\n{preview_buf}"
        elif result["type"] == "function":
            return result["description"]
        else:
            return result["value"]

    # response methods
    def _target_dispatch_message_from_target(self, response: dict):
        target_message = json.loads(response["params"]["message"])
        receive_message_id = target_message.get("id")
        if receive_message_id is None:
            self._missing_id_in_message(target_message)
            return
        self._dispatch_message_responses[receive_message_id] = response

    def _missing_id_in_message(self, message: dict):
        handler = self.response_methods.get(message["method"])
        if handler is not None:
            handler(message)
        else:
            logger.critical(f"unhandled message: {message}")

    def _console_message_added(self, message: dict):
        log_level = message["params"]["message"]["level"]
        text = message["params"]["message"]["text"]
        self._last_console_message = message
        webinspector_logger_handlers[log_level](text)

    def _console_message_repeated_count_updated(self, message: dict):
        self._console_message_added(self._last_console_message)

    def _heap_garbage_collected(self, message: dict):
        heap_logger.debug(message["params"])

    def _target_created(self, response: dict):
        pass

    def _target_destroyed(self, response: dict):
        pass

    def _target_did_commit_provisional_target(self, response: dict):
        self.set_target_id(response["params"]["newTargetId"])



import re
from datetime import datetime
from io import StringIO

import pytest

from pymobiledevice3.cli import syslog as syslog_module
from pymobiledevice3.services.os_trace import SyslogEntry, SyslogLogLevel

_FAKE_SERVICE_PROVIDER = object()


class _FakeOsTraceService:
    def __init__(self, entries):
        self._entries = entries

    async def syslog(self, pid=-1):
        for entry in self._entries:
            yield entry


def _create_syslog_entry(message: str) -> SyslogEntry:
    return SyslogEntry(
        pid=123,
        timestamp=datetime(2024, 1, 1, 0, 0, 0),
        level=SyslogLogLevel.INFO,
        image_name="/usr/libexec/test-process",
        image_offset=0,
        filename="/usr/libexec/test-process",
        message=message,
    )


def _create_syslog_entries(messages: list[str]) -> list[SyslogEntry]:
    return [_create_syslog_entry(message) for message in messages]


async def _run_syslog_live(
    monkeypatch, capsys, entries: list[SyslogEntry], **syslog_live_kwargs
) -> tuple[list[str], str]:
    monkeypatch.setattr(syslog_module, "OsTraceService", lambda lockdown: _FakeOsTraceService(entries))
    monkeypatch.setattr(syslog_module, "user_requested_colored_output", lambda: False)

    out = syslog_live_kwargs.pop("out", None)
    kwargs = {
        "pid": -1,
        "process_name": None,
        "match": [],
        "invert_match": [],
        "match_insensitive": [],
        "invert_match_insensitive": [],
        "include_label": False,
        "regex": [],
        "insensitive_regex": [],
    }
    kwargs.update(syslog_live_kwargs)

    await syslog_module.syslog_live(
        service_provider=_FAKE_SERVICE_PROVIDER,
        out=out,
        **kwargs,
    )

    printed_lines = capsys.readouterr().out.strip().splitlines()
    return printed_lines, "" if out is None else out.getvalue()


@pytest.mark.parametrize(
    ("invert_match", "invert_match_insensitive", "expected"),
    [
        ([], [], False),
        (["match"], [], True),
        (["MobileSafari"], [], True),
        (["dont"], [], False),
        ([], ["MaTCh"], True),
        ([], ["mobilesafari"], True),
        ([], ["missing"], False),
        (["match"], ["missing"], True),
        (["dont"], ["MaTCh"], True),
        (["dont"], ["missing"], False),
        (["match", "dont"], [], True),
        ([], ["missing", "MaTCh"], True),
        (["dont", "missing"], ["absent"], False),
    ],
)
def test_should_skip_line(invert_match: list[str], invert_match_insensitive: list[str], expected: bool) -> None:
    assert (
        syslog_module._should_skip_line(
            "MobileSafari match",
            invert_match,
            invert_match_insensitive,
        )
        == expected
    )


@pytest.mark.parametrize(
    ("match", "match_insensitive", "match_regex", "expected"),
    [
        ([], [], [], True),
        (["match"], [], [], True),
        (["MobileSafari"], [], [], True),
        (["missing"], [], [], False),
        (["MobileSafari", "match"], [], [], True),
        (["MobileSafari", "missing"], [], [], False),
        ([], ["mobilesafari"], [], True),
        ([], ["match"], [], True),
        ([], ["missing"], [], False),
        ([], ["mobilesafari", "match"], [], True),
        ([], ["mobilesafari", "missing"], [], False),
        (["match"], ["mobilesafari"], [], True),
        (["MobileSafari"], ["match"], [], True),
        (["match"], ["missing"], [], False),
        (["missing"], ["mobilesafari"], [], False),
        (["MobileSafari", "match"], ["mobilesafari", "match"], [], True),
        (["MobileSafari", "match"], ["mobilesafari", "missing"], [], False),
        ([], [], [syslog_module.re.compile(r".*(match).*")], True),
        ([], [], [syslog_module.re.compile(r".*(missing).*")], False),
        ([], [], [syslog_module.re.compile(r".*(match).*"), syslog_module.re.compile(r".*(missing).*")], True),
        (["match"], [], [syslog_module.re.compile(r".*(MobileSafari).*")], True),
        (["match"], [], [syslog_module.re.compile(r".*(missing).*")], False),
        ([], ["mobilesafari"], [syslog_module.re.compile(r".*(match).*")], True),
        ([], ["mobilesafari"], [syslog_module.re.compile(r".*(missing).*")], False),
    ],
)
def test_should_keep_line(
    match: list[str], match_insensitive: list[str], match_regex: list[re.Pattern[str]], expected: bool
) -> None:
    assert (
        syslog_module._should_keep_line(
            "MobileSafari match",
            match,
            match_insensitive,
            match_regex,
        )
        == expected
    )


@pytest.mark.asyncio
async def test_syslog_live_invert_match(monkeypatch, capsys):
    printed_lines, _ = await _run_syslog_live(
        monkeypatch,
        capsys,
        _create_syslog_entries(["keep this line", "skip this line"]),
        invert_match=["skip"],
    )
    assert len(printed_lines) == 1
    assert printed_lines[0].endswith("keep this line")


@pytest.mark.asyncio
async def test_syslog_live_match(monkeypatch, capsys):
    printed_lines, _ = await _run_syslog_live(
        monkeypatch,
        capsys,
        _create_syslog_entries(["daemon ready", "daemon error", "worker ready"]),
        match=["daemon"],
        invert_match=["error"],
    )
    assert len(printed_lines) == 1
    assert printed_lines[0].endswith("daemon ready")


@pytest.mark.asyncio
async def test_syslog_live_invert_match_uses_disjunction_for_repeated_values(monkeypatch, capsys):
    printed_lines, _ = await _run_syslog_live(
        monkeypatch,
        capsys,
        _create_syslog_entries(["daemon ready", "daemon ready error", "worker error"]),
        invert_match=["daemon", "error"],
    )
    assert len(printed_lines) == 0


@pytest.mark.asyncio
async def test_syslog_live_match_insensitive(monkeypatch, capsys):
    printed_lines, _ = await _run_syslog_live(
        monkeypatch,
        capsys,
        _create_syslog_entries(["MobileSafari ready", "backboardd ready", "Worker READY"]),
        match_insensitive=["mobilesafari"],
        invert_match_insensitive=["ready"],
    )
    assert len(printed_lines) == 0


@pytest.mark.asyncio
async def test_syslog_live_process_name_and_start_after_filter_output(monkeypatch, capsys):
    entries = [
        _create_syslog_entry("warmup"),
        SyslogEntry(
            pid=123,
            timestamp=datetime(2024, 1, 1, 0, 0, 1),
            level=SyslogLogLevel.INFO,
            image_name="/usr/libexec/other-process",
            image_offset=0,
            filename="/usr/libexec/other-process",
            message="START now",
        ),
        _create_syslog_entry("before START"),
        _create_syslog_entry("START now"),
        _create_syslog_entry("after start"),
    ]

    printed_lines, _ = await _run_syslog_live(
        monkeypatch,
        capsys,
        entries,
        process_name="test-process",
        start_after="START",
    )
    assert len(printed_lines) == 4
    assert printed_lines[0] == 'Waiting for "START" ...'
    assert printed_lines[1].endswith("before START")
    assert printed_lines[2].endswith("START now")
    assert printed_lines[3].endswith("after start")


@pytest.mark.asyncio
async def test_syslog_live_regex_filters_and_writes_plain_output_to_out(monkeypatch, capsys):
    entries = [
        _create_syslog_entry("daemon ready"),
        _create_syslog_entry("worker ready"),
        _create_syslog_entry("springboard ready"),
    ]
    out = StringIO()

    printed_lines, out_value = await _run_syslog_live(
        monkeypatch,
        capsys,
        entries,
        out=out,
        regex=["daemon", "worker"],
    )
    out_lines = out_value.strip().splitlines()
    assert len(printed_lines) == 2
    assert len(out_lines) == 2
    assert printed_lines[0].endswith("daemon ready")
    assert printed_lines[1].endswith("worker ready")
    assert out_lines == printed_lines



import contextlib
import logging
import plistlib
import tempfile
import traceback
from collections.abc import Iterator
from pathlib import Path
from typing import IO, Annotated, Optional, Union

import click
import requests
import typer
from ipsw_parser.ipsw import IPSW
from pygments import formatters, highlight, lexers
from typer_injector import Depends, InjectingTyper

from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import (
    async_command,
    cli_loop,
    is_invoked_for_completion,
    print_json,
    prompt_selection,
)
from pymobiledevice3.exceptions import ConnectionFailedError, ConnectionFailedToUsbmuxdError, IncorrectModeError
from pymobiledevice3.irecv import IRecv
from pymobiledevice3.lockdown import create_using_usbmux
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.recovery import Behavior, Recovery
from pymobiledevice3.restore.restore import Restore
from pymobiledevice3.services.diagnostics import DiagnosticsService
from pymobiledevice3.utils import file_download, start_ipython_shell

logger = logging.getLogger(__name__)


SHELL_USAGE = """
# use `irecv` variable to access Restore mode API
# for example:
print(irecv.getenv('build-version'))
"""
IPSWME_API = "https://api.ipsw.me/v4/device/"


cli = InjectingTyper(
    name="restore",
    help="Restore/erase IPSWs, fetch blobs, and manage devices in Recovery/DFU.",
    no_args_is_help=True,
)


async def _device_dependency_async(
    ecid: Annotated[
        Optional[str],
        typer.Option(
            help="Target device ECID; defaults to the first connected USB device or waits for Recovery/DFU.",
        ),
    ] = None,
) -> Optional[Device]:
    if is_invoked_for_completion():
        # prevent lockdown connection establishment when in autocomplete mode
        return None

    logger.debug("searching among connected devices via lockdownd")
    devices = [dev for dev in await usbmux.list_devices() if dev.connection_type == "USB"]
    if len(devices) > 1:
        raise click.ClickException("Multiple device detected")
    try:
        for device in devices:
            try:
                lockdown = await create_using_usbmux(serial=device.serial, connection_type="USB")
            except (ConnectionFailedError, IncorrectModeError):
                continue
            if (ecid is None) or (lockdown.ecid == ecid):
                logger.debug("found device")
                return Device(lockdown=lockdown)
            else:
                continue
    except ConnectionFailedToUsbmuxdError:
        pass

    logger.debug("waiting for device to be available in Recovery mode")
    return Device(irecv=IRecv(ecid=ecid))


def device_dependency(
    ecid: Annotated[
        Optional[str],
        typer.Option(
            help="Target device ECID; defaults to the first connected USB device or waits for Recovery/DFU.",
        ),
    ] = None,
) -> Optional[Device]:
    return cli_loop.run_until_complete(_device_dependency_async(ecid))


DeviceDep = Annotated[
    Device,
    Depends(device_dependency),
]


@contextlib.contextmanager
def tempzip_download_ctx(url: str) -> Iterator[IPSW]:
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpzip = Path(tmpdir) / url.split("/")[-1]
        file_download(url, tmpzip)
        with ipsw_ctx(tmpzip) as ipsw:
            yield ipsw


@contextlib.contextmanager
def ipsw_ctx(path: Union[str, Path]) -> Iterator[IPSW]:
    yield IPSW.create_from_path(path)


def ipsw_ctx_dependency(
    device: DeviceDep,
    ipsw: Annotated[
        Optional[str],
        typer.Option(
            "--ipsw",
            "-i",
            help="Path or URL to an IPSW. If omitted, choose a signed build interactively.",
        ),
    ] = None,
) -> contextlib.AbstractContextManager[IPSW]:
    if ipsw and not ipsw.startswith(("http://", "https://")):
        return ipsw_ctx(ipsw)

    url = ipsw
    if url is None:
        url = query_ipswme(cli_loop.run_until_complete(device.get_product_type()))
    return tempzip_download_ctx(url)


IPSWCtxDep = Annotated[
    contextlib.AbstractContextManager[IPSW],
    Depends(ipsw_ctx_dependency),
]


def tss_dependency(
    tss: Annotated[
        Optional[Path],
        typer.Option(help="Path to SHSH blob plist to use for signing requests."),
    ] = None,
) -> None:
    if tss is None:
        return
    with tss.open("rb") as tss_file:
        return plistlib.load(tss_file)


TSSDep = Annotated[
    Optional[dict],
    Depends(tss_dependency),
]


BehaviorOption = Annotated[
    Behavior,
    typer.Option(help="Restore behavior to use when selecting the BuildIdentity."),
]


def query_ipswme(identifier: str) -> str:
    resp = requests.get(IPSWME_API + identifier, headers={"Accept": "application/json"})
    firmwares = resp.json()["firmwares"]
    display_list = [f"{entry['version']}: {entry['buildid']}" for entry in firmwares if entry["signed"]]
    idx = prompt_selection(display_list, "Choose version", idx=True)
    return firmwares[idx]["url"]


async def restore_update_task(device: Device, ipsw: IPSW, tss: Optional[dict], erase: bool, ignore_fdr: bool) -> None:
    behavior = Behavior.Update
    if erase:
        behavior = Behavior.Erase

    try:
        await Restore(ipsw, device, tss=tss, behavior=behavior, ignore_fdr=ignore_fdr).update()
    except Exception:
        # click may "swallow" several exception types so we try to catch them all here
        traceback.print_exc()
        raise


@cli.command("shell")
def restore_shell(device: DeviceDep) -> None:
    """create an IPython shell for interacting with iBoot"""
    start_ipython_shell(
        header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.Terminal256Formatter(style="native")),
        user_ns={
            "irecv": device,
        },
    )


@cli.command("enter")
@async_command
async def restore_enter(device: DeviceDep) -> None:
    """enter Recovery mode"""
    if await device.get_is_lockdown():
        await device.lockdown.enter_recovery()


@cli.command("exit")
def restore_exit() -> None:
    """exit Recovery mode"""
    irecv = IRecv()
    irecv.set_autoboot(True)
    irecv.reboot()


@cli.command("restart")
@async_command
async def restore_restart(device: DeviceDep) -> None:
    """restarts device"""
    if await device.get_is_lockdown():
        async with DiagnosticsService(device.lockdown) as diagnostics:
            await diagnostics.restart()
    else:
        device.irecv.reboot()


async def restore_tss_task(
    device: Device,
    ipsw_ctx: contextlib.AbstractContextManager[IPSW],
    out: Optional[IO],
    behavior: Behavior = Behavior.Update,
) -> None:
    with ipsw_ctx as ipsw:
        tss = await Recovery(ipsw, device, behavior=behavior).fetch_tss_record()
    if out:
        plistlib.dump(tss, out)
    print_json(tss)


@cli.command("tss")
@async_command
async def restore_tss(
    device: DeviceDep,
    ipsw_ctx: IPSWCtxDep,
    out: Optional[Path] = None,
    behavior: BehaviorOption = Behavior.Update,
) -> None:
    """query SHSH blobs"""
    with out.open("wb") if out else contextlib.nullcontext() as out_file:
        await restore_tss_task(device, ipsw_ctx, out_file, behavior=behavior)


async def restore_ramdisk_task(device: Device, ipsw_ctx: contextlib.AbstractContextManager[IPSW]) -> None:
    with ipsw_ctx as ipsw:
        await Recovery(ipsw, device).boot_ramdisk()


@cli.command("ramdisk")
@async_command
async def restore_ramdisk(device: DeviceDep, ipsw_ctx: IPSWCtxDep) -> None:
    """
    Boot only the update ramdisk without performing a restore (IPSW path or URL accepted).
    """
    await restore_ramdisk_task(device, ipsw_ctx)


@cli.command("update")
@async_command
async def restore_update(
    device: DeviceDep,
    ipsw_ctx: IPSWCtxDep,
    tss: TSSDep,
    erase: Annotated[
        bool,
        typer.Option(help="Erase and restore (factory reset) instead of updating in place."),
    ] = False,
    ignore_fdr: Annotated[
        bool,
        typer.Option(help="Connect to the FDR service only (debug mode; no traffic proxying)."),
    ] = False,
) -> None:
    """
    Update or restore the device using an IPSW (local path or URL).
    """
    with ipsw_ctx as ipsw:
        await restore_update_task(device, ipsw, tss, erase, ignore_fdr)



import asyncio
from collections.abc import AsyncGenerator
from typing import Any

from pymobiledevice3.dtx import ConnectionAwareQueue, DTXService, dtx_method, dtx_on_dispatch, dtx_on_notification
from pymobiledevice3.dtx_service import DtxService


class GraphicsService(DTXService):
    IDENTIFIER = "com.apple.instruments.server.services.graphics.opengl"

    def __init__(self, ctx):
        super().__init__(ctx)
        self.events: asyncio.Queue[Any] = ConnectionAwareQueue()

    @dtx_method("startSamplingAtTimeInterval:")
    async def start_sampling_at_time_interval_(self, interval: float) -> Any: ...

    @dtx_method("stopSampling", expects_reply=False)
    async def stop_sampling(self) -> None: ...

    @dtx_on_dispatch
    async def _on_dispatch(self, selector: str, *args: Any) -> None:
        await self.events.put((selector, list(args)))

    @dtx_on_notification
    async def _on_notification(self, payload: Any) -> None:
        await self.events.put(payload)


class Graphics(DtxService[GraphicsService]):
    async def __aenter__(self):
        await self.connect()
        await self.service.start_sampling_at_time_interval_(0.0)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.service.stop_sampling()

    async def __aiter__(self) -> AsyncGenerator[Any, None]:
        while True:
            yield await self.service.events.get()



import logging
from typing import Annotated

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command
from pymobiledevice3.resources.firmware_notifications import get_notifications
from pymobiledevice3.services.notification_proxy import NotificationProxyService

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="notification",
    help="Post or observe Darwin notifications via notification_proxy.",
    no_args_is_help=True,
)


@cli.command()
@async_command
async def post(
    service_provider: ServiceProviderDep,
    names: list[str],
    insecure: Annotated[
        bool,
        typer.Option(help="Use the insecure relay meant for untrusted clients instead of the trusted channel."),
    ],
) -> None:
    """Post one or more Darwin notifications (notify_post)."""
    service = NotificationProxyService(lockdown=service_provider, insecure=insecure)
    for name in names:
        await service.notify_post(name)


@cli.command()
@async_command
async def observe(
    service_provider: ServiceProviderDep,
    names: list[str],
    insecure: Annotated[
        bool,
        typer.Option(help="Use the insecure relay meant for untrusted clients instead of the trusted channel."),
    ],
) -> None:
    """Subscribe and stream notifications (notify_register_dispatch)."""
    service = NotificationProxyService(lockdown=service_provider, insecure=insecure)
    for name in names:
        await service.notify_register_dispatch(name)

    async for event in service.receive_notification():
        logger.info(event)


@cli.command("observe-all")
@async_command
async def observe_all(
    service_provider: ServiceProviderDep,
    insecure: Annotated[
        bool,
        typer.Option(help="Use the insecure relay meant for untrusted clients instead of the trusted channel."),
    ],
) -> None:
    """Subscribe to all known firmware notifications and stream events."""
    service = NotificationProxyService(lockdown=service_provider, insecure=insecure)
    for notification in get_notifications():
        await service.notify_register_dispatch(notification)

    async for event in service.receive_notification():
        logger.info(event)



from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService


class IDAMService(LockdownService):
    RSD_SERVICE_NAME = "com.apple.idamd.shim.remote"
    SERVICE_NAME = "com.apple.idamd"

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    async def configuration_inquiry(self) -> dict:
        return await self.service.send_recv_plist({"Configuration Inquiry": True})

    async def set_idam_configuration(self, value: bool) -> None:
        await self.service.send_recv_plist({"Set IDAM Configuration": value})



#!/usr/bin/env python3

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService


class PreboardService(LockdownService):
    RSD_SERVICE_NAME = "com.apple.preboardservice_v2.shim.remote"
    SERVICE_NAME = "com.apple.preboardservice_v2"

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    async def create_stashbag(self, manifest):
        return await self.service.send_recv_plist({"Command": "CreateStashbag", "Manifest": manifest})

    async def commit(self, manifest):
        return await self.service.send_recv_plist({"Command": "CommitStashbag", "Manifest": manifest})



import asyncio
import uuid
from contextlib import asynccontextmanager

from fastapi import FastAPI, WebSocket
from fastapi.logger import logger

from pymobiledevice3.services.web_protocol.cdp_target import CdpTarget
from pymobiledevice3.services.web_protocol.session_protocol import SessionProtocol
from pymobiledevice3.services.webinspector import WirTypes


@asynccontextmanager
async def lifespan(app: FastAPI):
    await app.state.inspector.connect()
    yield


app = FastAPI(lifespan=lifespan)


@app.get("/json{_:path}")
async def available_targets(_: str):
    app.state.inspector.get_open_pages()
    targets = []
    for app_id in app.state.inspector.application_pages:
        for page_id, page in app.state.inspector.application_pages[app_id].items():
            if page.type_ not in (WirTypes.WEB, WirTypes.WEB_PAGE):
                continue
            targets.append({
                "description": "",
                "id": page_id,
                "title": page.web_title,
                "type": "page",
                "url": page.web_url,
                "webSocketDebuggerUrl": f"ws://localhost:9222/devtools/page/{page_id}",
                "devtoolsFrontendUrl": f"/devtools/inspector.html?ws://localhost:9222/devtools/page/{page_id}",
            })
    return targets


@app.get("/json/version")
def version():
    return {
        "Browser": "Safari",
        "Protocol-Version": "1.1",
        "User-Agent": "pymobiledevice3",
        "V8-Version": "7.2.233",
        "WebKit-Version": "537.36 (@cfede9db1d154de0468cb0538479f34c0755a0f4)",
        "webSocketDebuggerUrl": f"ws://localhost:9222/devtools/browser/{app.state.inspector.connection_id}",
    }


async def from_cdp(target: CdpTarget, websocket):
    async for message in websocket.iter_json():
        logger.debug(f"CDP INPUT:  {message}")
        await target.send(message)


async def to_cdp(target: CdpTarget, websocket):
    while True:
        message = await target.receive()
        logger.debug(f"CDP OUTPUT:  {message}")
        await websocket.send_json(message)


@app.websocket("/devtools/page/{page_id}")
async def page_debugger(websocket: WebSocket, page_id: str):
    application, page = app.state.inspector.find_page_id(page_id)
    session_id = str(uuid.uuid4()).upper()
    protocol = SessionProtocol(app.state.inspector, session_id, application, page, method_prefix="")
    target = await CdpTarget.create(protocol)
    await websocket.accept()
    await asyncio.gather(
        from_cdp(target, websocket),
        to_cdp(target, websocket),
    )



"""DTX service base classes and decorator machinery.

This module provides the building blocks for implementing DTX-based services:

Decorators
~~~~~~~~~~
- :func:`dtx_method`           — declare an outgoing ObjC selector call
- :func:`dtx_on_invoke`        — handle a specific incoming selector
- :func:`dtx_on_data`          — handle incoming DATA frames
- :func:`dtx_on_notification`  — handle incoming server-initiated OBJECT/OK
- :func:`dtx_on_dispatch`      — catch-all for unmatched incoming selectors

Service classes
~~~~~~~~~~~~~~~
- :class:`DTXService`          — base class; wires decorators to channel callbacks
- :class:`DTXDynamicService`   — dynamic proxy via ``__getattr__``
- :class:`DTXControlService`   — channel-0 capability handshake and channel lifecycle
- :class:`DTXProxyService`     — bidirectional proxy bridging two services on one channel

Helper functions
~~~~~~~~~~~~~~~~
- :func:`_python_name_to_objc_selector` — convert Python ``foo_bar_`` → ``foo:bar:``
- :func:`_objc_selector_to_python_name` — convert ``foo:bar:`` → ``foo_bar_``
- :func:`_apply_primitive_coercions`    — coerce args via annotation-driven type hints
"""

from __future__ import annotations

import inspect
import logging
import re
import sys
from asyncio import Queue
from collections.abc import Awaitable
from functools import partial, wraps
from typing import Any, Callable, ClassVar, Optional, TypeVar, get_type_hints

from pymobiledevice3.exceptions import ConnectionTerminatedError

from .channel import DTXChannel
from .context import DTX_GLOBAL_CTX, DTXContext  # noqa: F401 — re-exported for back-compat
from .ns_types import NSError
from .primitives import PInt32, _PrimitiveBase

logger = logging.getLogger(__name__)

_MISSING = object()  # sentinel for missing attribute values in __init_subclass__
DTX_SERVICE_T = TypeVar("DTX_SERVICE_T", bound="DTXService")
QUEUE_ITEM_T = TypeVar("QUEUE_ITEM_T")

__namespace_re__ = re.compile(r"^_([A-Z_]+_)+")


# ---------------------------------------------------------------------------
# Selector name conversion
# ---------------------------------------------------------------------------


def _objc_selector_to_python_name(selector: str) -> str:
    """Convert an Objective-C selector name to a valid Python attribute name.

    Colons in the selector become underscores::

        "foo:bar:" → "foo_bar_"
    """
    if not selector:
        return selector

    return selector.replace(":", "_")


def _python_name_to_objc_selector(name: str) -> str:
    """Convert a Python attribute name to an Objective-C selector name.

    Underscores become colons.  A leading ``_UPPERCASE_`` namespace prefix is
    preserved and its suffix is converted independently::

        "_XCT_logMessage_" → "_XCT_logMessage:"
        "runningProcesses"  → "runningProcesses"
        "setConfig_"        → "setConfig:"
    """
    if not name:
        return name

    match = __namespace_re__.match(name)
    if match:
        prefix = match.group(0)
        suffix = name[len(prefix) :]
        return prefix + suffix.replace("_", ":")
    elif name.startswith("_"):
        return "_" + name[1:].replace("_", ":")

    return name.replace("_", ":")


# ---------------------------------------------------------------------------
# Primitive coercion helper
# ---------------------------------------------------------------------------


def _apply_primitive_coercions(args: tuple, coercions: tuple) -> tuple:
    """Coerce *args* to :class:`_PrimitiveBase` types driven by annotation hints.

    For each position *i*, if ``coercions[i]`` is a :class:`_PrimitiveBase`
    subclass **and** the argument is not already a :class:`_PrimitiveBase`
    instance (so an explicit call-site type always wins), the argument is wrapped
    with the annotated type.

    Positions beyond ``len(coercions)`` are left unchanged (supports variadic
    calls).
    """
    result = list(args)
    for i, coerce_type in enumerate(coercions):
        if i >= len(result):
            break
        if coerce_type is None:
            continue
        if not (isinstance(coerce_type, type) and issubclass(coerce_type, _PrimitiveBase)):
            continue
        if not isinstance(result[i], _PrimitiveBase):
            result[i] = coerce_type(result[i])
    return tuple(result)


# ---------------------------------------------------------------------------
# DTX service decorators
# ---------------------------------------------------------------------------


def dtx_method(selector_or_fn=None, /, **invoke_kwargs):
    """Decorator for outgoing DTX calls.

    Replaces the decorated method body with a ``self._channel.invoke(...)``
    call generated at class-creation time.  Type annotations on stub
    parameters drive automatic :class:`_PrimitiveBase` coercion.

    Usage::

        @dtx_method                             # infer selector from Python name
        @dtx_method("setConfig:")               # explicit ObjC selector
        @dtx_method(expects_reply=False)        # inferred selector + fixed kwargs
        @dtx_method("setConfig:", expects_reply=False)
    """
    if callable(selector_or_fn):
        selector_or_fn._dtx_method = (None, {})
        return selector_or_fn
    selector = selector_or_fn

    def decorator(fn):
        fn._dtx_method = (selector, invoke_kwargs)
        return fn

    return decorator


def dtx_on_invoke(selector_or_fn=None, /):
    """Register a method as the handler for a specific incoming ObjC selector.

    Usage::

        @dtx_on_invoke                          # infer selector from Python name
        @dtx_on_invoke("_XCT_logMessage:")      # explicit selector
    """
    if callable(selector_or_fn):
        selector_or_fn._dtx_on_invoke = None  # None → infer from method name
        return selector_or_fn
    selector = selector_or_fn

    def decorator(fn):
        fn._dtx_on_invoke = selector
        return fn

    return decorator


def dtx_on_data(fn):
    """Register a method as the handler for incoming DATA frames (raw bytes)."""
    fn._dtx_on_data = True
    return fn


def dtx_on_notification(fn):
    """Register a method as the handler for incoming OBJECT/OK notifications."""
    fn._dtx_on_notification = True
    return fn


def dtx_on_dispatch(fn):
    """Catch-all handler for incoming DISPATCH messages not matched by
    :func:`dtx_on_invoke`.  The method receives ``(selector: str, *args)``."""
    fn._dtx_on_dispatch = True
    return fn


# ---------------------------------------------------------------------------
# DTXService
# ---------------------------------------------------------------------------


class ConnectionAwareQueue(Queue[QUEUE_ITEM_T]):
    """Queue that turns connection-termination sentinels back into exceptions."""

    async def get(self) -> QUEUE_ITEM_T:
        item = await super().get()
        if isinstance(item, ConnectionTerminatedError):
            raise item
        return item


class DTXService:
    """Base class for services communicating over a DTX channel.

    Subclasses declare outgoing calls with :func:`dtx_method` and incoming
    message handlers with :func:`dtx_on_invoke`, :func:`dtx_on_data`,
    :func:`dtx_on_notification`, or :func:`dtx_on_dispatch`.
    Channel callbacks are wired automatically in :meth:`__init__`.
    """

    IDENTIFIER: ClassVar[Optional[str]] = None
    _queue_attrs_on_close: ClassVar[tuple[str, ...]] = ("messages", "events", "output_events")

    # Class-level routing tables, populated by __init_subclass__.
    _dtx_dispatch: ClassVar[dict[str, str]] = {}
    _dtx_data_handler: ClassVar[Optional[str]] = None
    _dtx_notification_handler: ClassVar[Optional[str]] = None
    _dtx_dispatch_handler: ClassVar[Optional[str]] = None

    def __init_subclass__(cls, **kw: Any) -> None:
        super().__init_subclass__(**kw)

        new_dispatch: dict[str, str] = {}
        new_data_handler: Optional[str] = None
        new_notification_handler: Optional[str] = None
        new_dispatch_handler: Optional[str] = None

        for name, val in vars(cls).items():
            if not callable(val):
                continue

            sel = getattr(val, "_dtx_on_invoke", _MISSING)
            if sel is not _MISSING:
                new_dispatch[sel if sel is not None else _python_name_to_objc_selector(name)] = name

            if getattr(val, "_dtx_on_data", False):
                new_data_handler = name

            if getattr(val, "_dtx_on_notification", False):
                new_notification_handler = name

            if getattr(val, "_dtx_on_dispatch", False):
                new_dispatch_handler = name

            dtx = getattr(val, "_dtx_method", None)
            if dtx is not None:
                method_selector, invoke_kwargs = dtx
                if method_selector is None:
                    method_selector = _python_name_to_objc_selector(name)
                _sel, _kw = method_selector, dict(invoke_kwargs)

                # Build per-parameter coercion table from type annotations.
                # With 'from __future__ import annotations' all annotations are
                # strings; get_type_hints() evaluates them in the module's namespace.
                _coercions: tuple = ()
                try:
                    module_globals = vars(sys.modules.get(cls.__module__, None) or {})
                    hints = get_type_hints(val, globalns=module_globals)
                    params = [p for p in inspect.signature(val).parameters if p != "self"]
                    _coercions = tuple(hints.get(p) for p in params)
                except Exception:
                    pass

                async def _wrapper(
                    self,
                    *args: Any,
                    __sel: str = _sel,
                    __kw: dict = _kw,
                    __coercions: tuple = _coercions,
                    **extra: Any,
                ) -> Any:
                    if __coercions:
                        args = _apply_primitive_coercions(args, __coercions)
                    return await self._channel.invoke(__sel, *args, **{**__kw, **extra})

                wraps(val)(_wrapper)
                _wrapper.__qualname__ = f"{cls.__qualname__}.{name}"
                setattr(cls, name, _wrapper)

        # Merge dispatch table: base entries first so child selectors override.
        merged: dict[str, str] = {}
        for base in reversed(cls.__mro__[1:]):
            if base is object:
                continue
            merged.update(base.__dict__.get("_dtx_dispatch", {}))
        merged.update(new_dispatch)
        cls._dtx_dispatch = merged

        # Only override handler attrs on classes that explicitly declare them
        # (uninherited ClassVar on DTXService provides None as the default).
        if new_data_handler is not None:
            cls._dtx_data_handler = new_data_handler
        if new_notification_handler is not None:
            cls._dtx_notification_handler = new_notification_handler
        if new_dispatch_handler is not None:
            cls._dtx_dispatch_handler = new_dispatch_handler

    def __init__(self, ctx: DTXContext) -> None:
        self._ctx = ctx
        self._channel: DTXChannel = ctx["channel"]

        # Skip wiring when instantiated as a sub-service of a DTXProxyService.
        # The proxy itself handles dispatch routing via its own @dtx_on_dispatch.
        if "dtxproxy" in ctx:
            return

        cls = type(self)
        if cls._dtx_dispatch or cls._dtx_dispatch_handler is not None:
            self._channel.on_invoke = self.__on_dispatch__

        data_handler = cls._dtx_data_handler
        if data_handler is not None:
            self._channel.on_data = getattr(self, data_handler)

        notif_handler = cls._dtx_notification_handler
        if notif_handler is not None:
            self._channel.on_notification = getattr(self, notif_handler)

    async def __aenter__(self: DTX_SERVICE_T) -> DTX_SERVICE_T:
        await self._channel.__aenter__()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self._channel.__aexit__(exc_type, exc_val, exc_tb)

    def _on_connection_closed(self, exc: Optional[ConnectionTerminatedError] = None) -> None:
        """Notify queue-backed consumers that the DTX connection has closed.

        The default implementation pushes a :class:`ConnectionTerminatedError`
        sentinel into common service queues such as ``messages`` / ``events`` /
        ``output_events`` so consumers blocked on ``queue.get()`` wake up and
        raise immediately instead of blocking forever.
        """
        exc = exc or ConnectionTerminatedError("Connection closed")
        for attr_name in self._queue_attrs_on_close:
            queue = getattr(self, attr_name, None)
            if isinstance(queue, Queue):
                queue.put_nowait(exc)

    @property
    def dtxproxy(self) -> Optional[DTXProxyService]:
        """If this service is owned by a :class:`DTXProxyService`, return the proxy instance."""
        return self._ctx.get("dtxproxy", None)

    async def __on_dispatch__(self, selector: str, args: list[Any]) -> Any:
        """Route an incoming DISPATCH to the right @dtx_on_invoke handler,
        falling back to the @dtx_on_dispatch catch-all."""
        method_name = type(self)._dtx_dispatch.get(selector)
        if method_name is not None:
            return await getattr(self, method_name)(*args)
        dispatch_handler = type(self)._dtx_dispatch_handler
        if dispatch_handler is not None:
            return await getattr(self, dispatch_handler)(selector, *args)
        logger.warning(
            "%s: received dispatch for selector %r with no registered handler",
            type(self).__name__,
            selector,
        )
        return NSError.create_doesnt_respond_to_selector(selector)

    # ------------------------------------------------------------------
    # Convenience wrappers for outgoing messages (usable without @dtx_method)
    # ------------------------------------------------------------------

    async def send_notification(self, notification: Any, *aux_args: Any, expects_reply: bool = False) -> Any:
        """Send a server-bound notification and optionally await the reply."""
        return await self._channel.notify(notification, *aux_args, expects_reply=expects_reply)

    async def invoke(self, method: str, *args: Any, expects_reply: bool = True) -> Any:
        """Invoke *method* on this channel and return the decoded reply value.

        This is a high-level convenience wrapper around :meth:`DTXChannel.invoke`.
        Prefer :func:`dtx_method`-decorated stubs for typed outgoing calls; use
        this method for ad-hoc or dynamic selector invocations.
        """
        return await self._channel.invoke(method, *args, expects_reply=expects_reply)

    async def do_invoke(self, method: str, *args: Any, expects_reply: bool = True) -> Any:
        """Backward-compatible alias for :meth:`invoke`."""
        return await self.invoke(method, *args, expects_reply=expects_reply)

    async def send_data(self, data: bytes, *aux_args: Any, expects_reply: bool = False) -> Any:
        """Send a DATA frame and optionally await the reply."""
        return await self._channel.send_data(data, *aux_args, expects_reply=expects_reply)


class DTXDynamicService(DTXService):
    """Dynamic proxy that maps attribute access to DTX selector calls.

    Any attribute not defined as a class method is routed through
    ``__getattr__`` → ``__getitem__`` → ``invoke``, translating Python
    names to ObjC selectors via :func:`_python_name_to_objc_selector`.

    Incoming dispatches are handled by :meth:`__dynamic_dispatch__`: for each
    incoming selector the corresponding Python-mangled name is looked up in
    subclass ``__dict__`` entries above ``DTXDynamicService`` in the MRO.
    """

    def __getitem__(self, item: str) -> Callable[..., Awaitable[Any]]:
        """Return a coroutine factory that invokes the given ObjC selector."""
        return partial(self.invoke, item)

    def __getattr__(self, name: str) -> Callable[..., Awaitable[Any]]:
        """Translate a Python attribute name to an ObjC selector and return a callable."""
        return self[_python_name_to_objc_selector(name)]

    @dtx_on_dispatch
    async def __dynamic_dispatch__(self, selector: str, *args: Any) -> Any:
        method = _objc_selector_to_python_name(selector)
        for cls in type(self).__mro__:
            if cls is DTXDynamicService:
                break
            if method in cls.__dict__:
                attr = getattr(self, method)
                if not callable(attr):
                    logger.warning(
                        "%s: dispatch for %r (method %r) is not callable",
                        type(self).__name__,
                        selector,
                        method,
                    )
                    return NSError.create_doesnt_respond_to_selector(selector)
                if inspect.iscoroutinefunction(attr):
                    return await attr(*args)
                return attr(*args)
        logger.warning(
            "%s: dispatch for %r (method %r) is not implemented",
            type(self).__name__,
            selector,
            method,
        )
        return NSError.create_doesnt_respond_to_selector(selector)


class DTXControlService(DTXService):
    """Service for DTX channel 0: capability handshake and channel lifecycle.

    Uses :func:`dtx_method` for the three outgoing control calls and
    :func:`dtx_on_invoke` for the three peer-initiated dispatches.
    Annotation-driven coercion (:class:`PInt32`) ensures channel codes are
    sent as INT32 without manual :class:`PInt32` wrapping.
    """

    # ------------------------------------------------------------------
    # Outgoing
    # ------------------------------------------------------------------

    @dtx_method("_notifyOfPublishedCapabilities:", expects_reply=False)
    async def notify_capabilities(self, capabilities: dict) -> None:
        """Announce our capability dictionary to the peer."""

    @dtx_method("_requestChannelWithCode:identifier:")
    async def request_channel(self, channel_code: PInt32, identifier: str) -> None:
        """Ask the peer to open the named service channel with the given code."""

    @dtx_method("_channelCanceled:")
    async def cancel_channel(self, channel_code: PInt32) -> None:
        """Tell the peer to cancel the channel with the given code."""

    # ------------------------------------------------------------------
    # Incoming
    # ------------------------------------------------------------------

    @dtx_on_invoke("_notifyOfPublishedCapabilities:")
    async def _recv_capabilities(self, capabilities: dict) -> None:
        await self._ctx["connection"]._on_capabilities_received(capabilities)

    @dtx_on_invoke("_requestChannelWithCode:identifier:")
    async def _recv_channel_request(self, code: int, identifier: str) -> Optional[str]:
        return await self._ctx["connection"]._on_channel_request(code, identifier)

    @dtx_on_invoke("_channelCanceled:")
    async def _recv_channel_cancelled(self, code: int) -> None:
        await self._ctx["connection"]._on_channel_cancelled(code)

    @dtx_on_dispatch
    async def _recv_unknown_control_dispatch(self, selector: str, *args: Any) -> None:
        queue = self._ctx["connection"].ctx.get("control_dispatch_queue")
        if queue is not None:
            await queue.put((selector, list(args)))
        logger.debug("ignoring unknown control-channel selector %r with %d arg(s)", selector, len(args))


class DTXProxyService(DTXService):
    """Bidirectional proxy that bridges two :class:`DTXService` instances on one channel.

    Incoming dispatches are forwarded to *local_service*; outgoing calls
    (``send_notification``, ``invoke``, ``send_data``) are forwarded to
    *remote_service*.

    Sub-services are assigned via the :attr:`local_service` and
    :attr:`remote_service` property setters, which also wire the channel
    callbacks.  Sub-services receive a child context with ``"dtxproxy"`` set
    to this proxy instance so their ``__init__`` skips callback wiring and
    they can reach their counterpart via ``self._ctx["dtxproxy"].remote_service``.

    Usage::

        proxy = DTXProxyService(ctx)
        proxy.local_service = IDEInterface(ctx.child(dtxproxy=proxy))
        proxy.remote_service = DaemonInterface(ctx.child(dtxproxy=proxy))
    """

    def __init__(self, ctx: DTXContext) -> None:
        super().__init__(ctx)
        self._local_service: Optional[DTXService] = None
        self._remote_service: Optional[DTXService] = None

    @property
    def local_service(self) -> DTXService:
        """The service that receives incoming dispatches from the remote end."""
        if self._local_service is None:
            raise AttributeError("local_service has not been assigned yet")
        return self._local_service

    @local_service.setter
    def local_service(self, svc: DTXService) -> None:
        self._local_service = svc
        # Wire the channel's dispatch/data callbacks to this proxy's handlers,
        # which delegate to the local service.  Re-wire on every assignment so
        # the proxy always reflects the current local service.
        cls = type(self)
        if cls._dtx_dispatch or cls._dtx_dispatch_handler is not None:
            self._channel.on_invoke = self.__on_dispatch__
        data_handler = cls._dtx_data_handler
        if data_handler is not None:
            self._channel.on_data = getattr(self, data_handler)

    @property
    def remote_service(self) -> DTXService:
        """The service whose channel methods are used for outgoing calls."""
        if self._remote_service is None:
            raise AttributeError("remote_service has not been assigned yet")
        return self._remote_service

    @remote_service.setter
    def remote_service(self, svc: DTXService) -> None:
        self._remote_service = svc
        notif_handler = type(self)._dtx_notification_handler
        if notif_handler is not None:
            self._channel.on_notification = getattr(self, notif_handler)

    @dtx_on_dispatch
    async def __proxy_dispatch__(self, selector: str, *args: Any) -> Any:
        return await self.local_service.__on_dispatch__(selector, list(args))

    @dtx_on_data
    async def __proxy_data__(self, data: bytes) -> Any:
        handler = type(self.local_service)._dtx_data_handler
        if handler:
            return await getattr(self.local_service, handler)(data)
        return None

    @dtx_on_notification
    async def __proxy_notification__(self, notification: Any) -> Any:
        handler = type(self.remote_service)._dtx_notification_handler
        if handler:
            return await getattr(self.remote_service, handler)(notification)
        return None

    async def send_notification(self, notification: Any, *aux_args: Any, expects_reply: bool = False) -> Any:
        """Forward a notification through the remote service."""
        return await self.remote_service.send_notification(notification, *aux_args, expects_reply=expects_reply)

    async def invoke(self, method: str, *args: Any, expects_reply: bool = True) -> Any:
        """Forward a selector invocation through the remote service."""
        return await self.remote_service.invoke(method, *args, expects_reply=expects_reply)

    async def send_data(self, data: bytes, *aux_args: Any, expects_reply: bool = False) -> Any:
        """Forward a DATA frame through the remote service."""
        return await self.remote_service.send_data(data, *aux_args, expects_reply=expects_reply)



__all__ = [
    "AccessDeniedError",
    "AfcException",
    "AfcFileNotFoundError",
    "AlreadyMountedError",
    "AmfiError",
    "ArbitrationError",
    "ArgumentError",
    "BadCommandError",
    "BadDevError",
    "CannotStopSessionError",
    "CloudConfigurationAlreadyPresentError",
    "ConnectionFailedError",
    "ConnectionFailedToUsbmuxdError",
    "ConnectionTerminatedError",
    "CoreDeviceError",
    "DeprecationError",
    "DeveloperModeError",
    "DeveloperModeIsNotEnabledError",
    "DeviceAlreadyInUseError",
    "DeviceHasPasscodeSetError",
    "DeviceNotFoundError",
    "DeviceVersionNotSupportedError",
    "DisableMemoryLimitError",
    "DvtDirListError",
    "DvtException",
    "ExtractingStackshotError",
    "FatalPairingError",
    "FeatureNotSupportedError",
    "GetProhibitedError",
    "IRecvError",
    "IRecvNoDeviceConnectedError",
    "IncorrectModeError",
    "InspectorEvaluateError",
    "InternalError",
    "InvalidConnectionError",
    "InvalidHostIDError",
    "InvalidServiceError",
    "LaunchingApplicationError",
    "LockdownError",
    "MessageNotSupportedError",
    "MissingValueError",
    "MuxException",
    "MuxVersionError",
    "NoDeviceConnectedError",
    "NotEnoughDiskSpaceError",
    "NotMountedError",
    "NotPairedError",
    "NotTrustedError",
    "NotificationTimeoutError",
    "OSNotSupportedError",
    "PairingDialogResponsePendingError",
    "PairingError",
    "PasscodeRequiredError",
    "PasswordRequiredError",
    "ProfileError",
    "PyMobileDevice3Exception",
    "QuicProtocolNotSupportedError",
    "RSDRequiredError",
    "RemoteAutomationNotEnabledError",
    "RemotePairingCompletedError",
    "SetProhibitedError",
    "StartServiceError",
    "SysdiagnoseTimeoutError",
    "TSSError",
    "TunneldConnectionError",
    "UnrecognizedSelectorError",
    "UnsupportedCommandError",
    "UserDeniedPairingError",
    "WdaError",
    "WebInspectorNotEnabledError",
    "WirError",
]

from typing import Optional


class PyMobileDevice3Exception(Exception):
    pass


class DeviceVersionNotSupportedError(PyMobileDevice3Exception):
    pass


class IncorrectModeError(PyMobileDevice3Exception):
    pass


class NotTrustedError(PyMobileDevice3Exception):
    pass


class PairingError(PyMobileDevice3Exception):
    pass


class NotPairedError(PyMobileDevice3Exception):
    pass


class CannotStopSessionError(PyMobileDevice3Exception):
    pass


class PasswordRequiredError(PairingError):
    pass


class StartServiceError(PyMobileDevice3Exception):
    def __init__(self, service_name: str, message: str) -> None:
        super().__init__()
        self.service_name = service_name
        self.message = message


class FatalPairingError(PyMobileDevice3Exception):
    pass


class NoDeviceConnectedError(PyMobileDevice3Exception):
    pass


class InterfaceIndexNotFoundError(PyMobileDevice3Exception):
    def __init__(self, address: str):
        super().__init__()
        self.address = address


class DeviceNotFoundError(PyMobileDevice3Exception):
    def __init__(self, udid: str):
        super().__init__()
        self.udid = udid


class TunneldConnectionError(PyMobileDevice3Exception):
    pass


class MuxException(PyMobileDevice3Exception):
    pass


class MuxVersionError(MuxException):
    pass


class BadCommandError(MuxException):
    pass


class BadDevError(MuxException):
    pass


class ConnectionFailedError(MuxException):
    pass


class ConnectionFailedToUsbmuxdError(ConnectionFailedError):
    pass


class ArgumentError(PyMobileDevice3Exception):
    pass


class AfcException(PyMobileDevice3Exception, OSError):
    def __init__(self, message: str, status: str, filename: Optional[str] = None) -> None:
        OSError.__init__(self, status, message)
        self.status = status
        self.filename = filename


class AfcFileNotFoundError(AfcException):
    pass


class DvtException(PyMobileDevice3Exception):
    """Domain exception for DVT operations."""

    pass


class UnrecognizedSelectorError(DvtException):
    """Attempted to call an unrecognized selector from DVT."""

    pass


class DvtDirListError(DvtException):
    """Raise when directory listing fails."""

    pass


class NotMountedError(PyMobileDevice3Exception):
    """Given image for umount wasn't mounted in the first place"""

    pass


class AlreadyMountedError(PyMobileDevice3Exception):
    """Given image for mount has already been mounted in the first place"""

    pass


class MissingManifestError(PyMobileDevice3Exception):
    """No manifest could be found"""

    pass


class UnsupportedCommandError(PyMobileDevice3Exception):
    """Given command isn't supported for this iOS version"""

    pass


class ExtractingStackshotError(PyMobileDevice3Exception):
    """Raise when stackshot is not received in the core profile session."""

    pass


class ChannelClosedError(PyMobileDevice3Exception):
    """Raised when trying to send or read a message on a closed channel."""

    pass


class ConnectionTerminatedError(PyMobileDevice3Exception):
    """Raise when a connection is terminated abruptly."""

    pass


class StreamClosedError(ConnectionTerminatedError):
    """Raise when trying to send a message on a closed stream."""

    pass


class WebInspectorNotEnabledError(PyMobileDevice3Exception):
    """Raise when Web Inspector is not enabled."""

    pass


class RemoteAutomationNotEnabledError(PyMobileDevice3Exception):
    """Raise when Web Inspector remote automation is not enabled."""

    pass


class WirError(PyMobileDevice3Exception):
    """Raise when Webinspector WIR command fails."""

    pass


class InternalError(PyMobileDevice3Exception):
    """Some internal Apple error"""

    pass


class ArbitrationError(PyMobileDevice3Exception):
    """Arbitration failed"""

    pass


class DeviceAlreadyInUseError(ArbitrationError):
    """Device is already checked-in by someone"""

    @property
    def message(self):
        return self.args[0].get("message")

    @property
    def owner(self):
        return self.args[0].get("owner")

    @property
    def result(self):
        return self.args[0].get("result")


class DeveloperModeIsNotEnabledError(PyMobileDevice3Exception):
    """Raise when mounting failed because developer mode is not enabled."""

    pass


class DeveloperDiskImageNotFoundError(PyMobileDevice3Exception):
    """Failed to locate the correct DeveloperDiskImage.dmg"""

    pass


class DeveloperModeError(PyMobileDevice3Exception):
    """Raise when amfid failed to enable developer mode."""

    pass


class LockdownError(PyMobileDevice3Exception):
    """lockdown general error"""

    def __init__(self, message: str, identifier: Optional[str] = None) -> None:
        super().__init__(message)
        self.identifier = identifier


class GetProhibitedError(LockdownError):
    pass


class SetProhibitedError(LockdownError):
    pass


class PairingDialogResponsePendingError(PairingError):
    """User hasn't yet confirmed the device is trusted"""

    pass


class UserDeniedPairingError(PairingError):
    pass


class InvalidHostIDError(PairingError):
    pass


class MissingValueError(LockdownError):
    """raised when attempting to query non-existent domain/key"""

    pass


class InvalidConnectionError(LockdownError):
    pass


class PasscodeRequiredError(LockdownError):
    """passcode must be present for this action"""

    pass


class AmfiError(PyMobileDevice3Exception):
    pass


class DeviceHasPasscodeSetError(AmfiError):
    pass


class NotificationTimeoutError(PyMobileDevice3Exception, TimeoutError):
    pass


class ProfileError(PyMobileDevice3Exception):
    pass


class CloudConfigurationAlreadyPresentError(ProfileError):
    pass


class IRecvError(PyMobileDevice3Exception):
    pass


class IRecvNoDeviceConnectedError(IRecvError):
    pass


class MessageNotSupportedError(PyMobileDevice3Exception):
    pass


class InvalidServiceError(LockdownError):
    pass


class InspectorEvaluateError(PyMobileDevice3Exception):
    def __init__(
        self,
        class_name: str,
        message: str,
        line: Optional[int] = None,
        column: Optional[int] = None,
        stack: Optional[list[str]] = None,
    ):
        super().__init__()
        self.class_name = class_name
        self.message = message
        self.line = line
        self.column = column
        self.stack = stack

    def __str__(self) -> str:
        stack_trace = "\n".join([f"\t - {frame}" for frame in self.stack])
        return f"{self.class_name}: {self.message}.\nLine: {self.line} Column: {self.column}\nStack: {stack_trace}"


class LaunchingApplicationError(PyMobileDevice3Exception):
    pass


class AppInstallError(PyMobileDevice3Exception):
    pass


class AppNotInstalledError(PyMobileDevice3Exception):
    pass


class CoreDeviceError(PyMobileDevice3Exception):
    pass


class AccessDeniedError(PyMobileDevice3Exception):
    """Need extra permissions to execute this command"""

    pass


class NoSuchBuildIdentityError(PyMobileDevice3Exception):
    pass


class MobileActivationException(PyMobileDevice3Exception):
    """Mobile activation can not be done"""

    pass


class NotEnoughDiskSpaceError(PyMobileDevice3Exception):
    """Computer does not have enough disk space for the intended operation"""

    pass


class DeprecationError(PyMobileDevice3Exception):
    """The requested action/service/method is deprecated"""

    pass


class RSDRequiredError(PyMobileDevice3Exception):
    """The requested action requires an RSD object"""

    def __init__(self, identifier: str) -> None:
        self.identifier = identifier
        super().__init__()


class SysdiagnoseTimeoutError(PyMobileDevice3Exception, TimeoutError):
    """Timeout collecting new sysdiagnose archive"""

    pass


class SupportError(PyMobileDevice3Exception):
    def __init__(self, os_name):
        self.os_name = os_name
        super().__init__()


class OSNotSupportedError(SupportError):
    """Operating system is not supported."""

    pass


class FeatureNotSupportedError(SupportError):
    """Feature has not been implemented for OS."""

    def __init__(self, os_name, feature):
        super().__init__(os_name)
        self.feature = feature


class QuicProtocolNotSupportedError(PyMobileDevice3Exception):
    """QUIC tunnel support was removed on iOS 18.2+"""

    pass


class RemotePairingCompletedError(PyMobileDevice3Exception):
    """
    Raised upon pairing completion using the `remotepairingdeviced` service (RemoteXPC).

    remotepairingdeviced closes connection after pairing, so client must re-establish it after pairing is
    completed.
    """

    pass


class DisableMemoryLimitError(PyMobileDevice3Exception):
    """Disabling memory limit fails."""

    pass


class ProtocolError(PyMobileDevice3Exception):
    """An unexpected protocol message was received"""

    pass


class TSSError(PyMobileDevice3Exception):
    """An unexpected message was received from apple ticket server"""

    pass


class WdaError(PyMobileDevice3Exception):
    """Raised for WebDriverAgent (WDA) errors."""

    def __init__(self, message: str, status_code: Optional[int] = None):
        super().__init__(message)
        self.status_code = status_code



# mbn.py
# Support for Qualcomm MBN (Modem Binary) formats — Python port
# Mirrors the logic of mbn.c (v1/v2/BIN headers, ELF detection, and v7 stitching)
#
# Copyright (c) 2012 Martin Szulecki
# Copyright (c) 2012 Nikias Bassen
# Copyright (c) 2025 Visual Ehrmanntraut <visual@chefkiss.dev>
#
# Ported to Python by DoronZ <doron88@gmail.com>. Licensed under LGPL-2.1-or-later (same as original).

import io
import logging
from typing import Optional

from construct import Bytes, ChecksumError, Int16ul, Int32ul, Int64ul, Struct

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------
MBN_V1_MAGIC = b"\x0a\x00\x00\x00"
MBN_V1_MAGIC_SIZE = 4

MBN_V2_MAGIC = b"\xd1\xdc\x4b\x84\x34\x10\xd7\x73"
MBN_V2_MAGIC_SIZE = 8

MBN_BIN_MAGIC = b"\x04\x00\xea\x6c\x69\x48\x55"
MBN_BIN_MAGIC_SIZE = 7
MBN_BIN_MAGIC_OFFSET = 1

# ELF
EI_MAG0, EI_MAG1, EI_MAG2, EI_MAG3, EI_CLASS = 0, 1, 2, 3, 4
EI_NIDENT = 16
ELFMAG0, ELFMAG1, ELFMAG2, ELFMAG3 = 0x7F, ord("E"), ord("L"), ord("F")
ELFCLASSNONE, ELFCLASS32, ELFCLASS64 = 0, 1, 2

# -----------------------------------------------------------------------------
# Construct Structs (little-endian)
# -----------------------------------------------------------------------------

# MBN v1
MBN_V1 = Struct(
    "type" / Int32ul,
    "unk_0x04" / Int32ul,
    "unk_0x08" / Int32ul,
    "unk_0x0c" / Int32ul,
    "data_size" / Int32ul,  # total - sizeof(header)
    "sig_offset" / Int32ul,  # real offset = enc_sig_offset & 0xFFFFFF00 (FYI)
    "unk_0x18" / Int32ul,
    "unk_0x1c" / Int32ul,
    "unk_0x20" / Int32ul,
    "unk_0x24" / Int32ul,
)

# MBN v2
MBN_V2 = Struct(
    "magic1" / Bytes(8),
    "unk_0x08" / Int32ul,
    "unk_0x0c" / Int32ul,  # 0xFFFFFFFF
    "unk_0x10" / Int32ul,  # 0xFFFFFFFF
    "header_size" / Int32ul,
    "unk_0x18" / Int32ul,
    "data_size" / Int32ul,  # total - sizeof(header)
    "sig_offset" / Int32ul,
    "unk_0x24" / Int32ul,
    "unk_0x28" / Int32ul,
    "unk_0x2c" / Int32ul,
    "unk_0x30" / Int32ul,
    "unk_0x34" / Int32ul,  # 0x1
    "unk_0x38" / Int32ul,  # 0x1
    "unk_0x3c" / Int32ul,  # 0xFFFFFFFF
    "unk_0x40" / Int32ul,  # 0xFFFFFFFF
    "unk_0x44" / Int32ul,  # 0xFFFFFFFF
    "unk_0x48" / Int32ul,  # 0xFFFFFFFF
    "unk_0x4c" / Int32ul,  # 0xFFFFFFFF
)

# MBN BIN
MBN_BIN = Struct(
    "magic" / Bytes(8),
    "unk_0x08" / Int32ul,
    "version" / Int32ul,
    "total_size" / Int32ul,  # includes header
    "unk_0x14" / Int32ul,
)

# v7 trailer header used by mbn_mav25_stitch
MBN_V7 = Struct(
    "reserved" / Int32ul,
    "version" / Int32ul,
    "common_metadata_size" / Int32ul,
    "qti_metadata_size" / Int32ul,
    "oem_metadata_size" / Int32ul,
    "hash_table_size" / Int32ul,
    "qti_signature_size" / Int32ul,
    "qti_certificate_chain_size" / Int32ul,
    "oem_signature_size" / Int32ul,
    "oem_certificate_chain_size" / Int32ul,
)

# ELF (minimal fields we need)
ELF32_Ehdr = Struct(
    "e_ident" / Bytes(16),
    "e_type" / Int16ul,
    "e_machine" / Int16ul,
    "e_version" / Int32ul,
    "e_entry" / Int32ul,
    "e_phoff" / Int32ul,
    "e_shoff" / Int32ul,
    "e_flags" / Int32ul,
    "e_ehsize" / Int16ul,
    "e_phentsize" / Int16ul,
    "e_phnum" / Int16ul,
    "e_shentsize" / Int16ul,
    "e_shnum" / Int16ul,
    "e_shstrndx" / Int16ul,
)

ELF64_Ehdr = Struct(
    "e_ident" / Bytes(16),
    "e_type" / Int16ul,
    "e_machine" / Int16ul,
    "e_version" / Int32ul,
    "e_entry" / Int64ul,
    "e_phoff" / Int64ul,
    "e_shoff" / Int64ul,
    "e_flags" / Int32ul,
    "e_ehsize" / Int16ul,
    "e_phentsize" / Int16ul,
    "e_phnum" / Int16ul,
    "e_shentsize" / Int16ul,
    "e_shnum" / Int16ul,
    "e_shstrndx" / Int16ul,
)

ELF32_Phdr = Struct(
    "p_type" / Int32ul,
    "p_offset" / Int32ul,
    "p_vaddr" / Int32ul,
    "p_paddr" / Int32ul,
    "p_filesz" / Int32ul,
    "p_memsz" / Int32ul,
    "p_flags" / Int32ul,
    "p_align" / Int32ul,
)

ELF64_Phdr = Struct(
    "p_type" / Int32ul,
    "p_flags" / Int32ul,
    "p_offset" / Int64ul,
    "p_vaddr" / Int64ul,
    "p_paddr" / Int64ul,
    "p_filesz" / Int64ul,
    "p_memsz" / Int64ul,
    "p_align" / Int64ul,
)


# -----------------------------------------------------------------------------
# Helpers
# -----------------------------------------------------------------------------


def _is_valid_elf_ident(e_ident: bytes) -> bool:
    return (
        len(e_ident) >= EI_NIDENT
        and e_ident[EI_MAG0] == ELFMAG0
        and e_ident[EI_MAG1] == ELFMAG1
        and e_ident[EI_MAG2] == ELFMAG2
        and e_ident[EI_MAG3] == ELFMAG3
        and e_ident[EI_CLASS] != ELFCLASSNONE
    )


def _read_elf_headers(data: bytes):
    if len(data) < EI_NIDENT:
        return None, None
    e_ident = data[:EI_NIDENT]
    if not _is_valid_elf_ident(e_ident):
        return None, None
    if e_ident[EI_CLASS] == ELFCLASS64:
        if len(data) < ELF64_Ehdr.sizeof():
            return None, None
        hdr = ELF64_Ehdr.parse(data[: ELF64_Ehdr.sizeof()])
        return "ELF64", hdr
    elif e_ident[EI_CLASS] == ELFCLASS32:
        if len(data) < ELF32_Ehdr.sizeof():
            return None, None
        hdr = ELF32_Ehdr.parse(data[: ELF32_Ehdr.sizeof()])
        return "ELF32", hdr
    return None, None


def _read_program_headers(data: bytes, kind: str, hdr) -> list:
    phdrs = []
    if hdr.e_phnum == 0:
        logger.error("%s: ELF has no program sections", "_read_program_headers")
        return phdrs

    table_size = hdr.e_phnum * (ELF64_Phdr.sizeof() if kind == "ELF64" else ELF32_Phdr.sizeof())
    if hdr.e_phoff + table_size > len(data):
        logger.error("%s: Program header table is out of bounds", "_read_program_headers")
        return []

    table = data[hdr.e_phoff : hdr.e_phoff + table_size]
    bio = io.BytesIO(table)
    for _ in range(hdr.e_phnum):
        ph = ELF64_Phdr.parse_stream(bio) if kind == "ELF64" else ELF32_Phdr.parse_stream(bio)
        phdrs.append(ph)
    return phdrs


def _elf_last_segment_end(data: bytes) -> Optional[int]:
    kind, hdr = _read_elf_headers(data)
    if not hdr:
        return None
    phdrs = _read_program_headers(data, kind, hdr)
    if not phdrs:
        return None
    # last by highest p_offset
    last = max(phdrs, key=lambda p: int(p.p_offset))
    return int(last.p_offset + last.p_filesz)


def _mbn_v7_header_sizes_valid(h, sect_size: int) -> bool:
    total = (
        MBN_V7.sizeof()
        + h.common_metadata_size
        + h.qti_metadata_size
        + h.oem_metadata_size
        + h.hash_table_size
        + h.qti_signature_size
        + h.qti_certificate_chain_size
        + h.oem_signature_size
        + h.oem_certificate_chain_size
    )
    return total <= sect_size


def _mbn_v7_header_sizes_expected(h) -> bool:
    return (
        (h.qti_metadata_size in (0, 0xE0))
        and (h.oem_metadata_size in (0, 0xE0))
        and (h.oem_signature_size in (0, 0x68))
        and (h.oem_certificate_chain_size in (0, 0xD20))
    )


def _mbn_v7_log(h, func: str, prefix: str) -> None:
    logger.debug(
        "%s: %s header {version=0x%x, common_metadata_size=0x%x, qti_metadata_size=0x%x, "
        "oem_metadata_size=0x%x, hash_table_size=0x%x, qti_signature_size=0x%x, "
        "qti_certificate_chain_size=0x%x, oem_signature_size=0x%x, oem_certificate_chain_size=0x%x}",
        func,
        prefix,
        h.version,
        h.common_metadata_size,
        h.qti_metadata_size,
        h.oem_metadata_size,
        h.hash_table_size,
        h.qti_signature_size,
        h.qti_certificate_chain_size,
        h.oem_signature_size,
        h.oem_certificate_chain_size,
    )


# -----------------------------------------------------------------------------
# Public API
# -----------------------------------------------------------------------------


def mbn_stitch(data: bytes, blob: bytes) -> Optional[bytes]:
    """
    Overwrite the tail of `data` with `blob`. Format-aware size logging/checks.
    Returns new bytes or None.
    """
    if data is None:
        logger.error("%s: data is NULL", "mbn_stitch")
        return None
    if not data:
        logger.error("%s: data size is 0", "mbn_stitch")
        return None
    if blob is None:
        logger.error("%s: blob is NULL", "mbn_stitch")
        return None
    if not blob:
        logger.error("%s: blob size is 0", "mbn_stitch")
        return None

    data_size = len(data)
    blob_size = len(blob)
    parsed_size = 0

    try:
        # MBN v2
        if data_size > MBN_V2_MAGIC_SIZE and data[:MBN_V2_MAGIC_SIZE] == MBN_V2_MAGIC:
            if data_size < MBN_V2.sizeof():
                logger.error("%s: truncated MBN v2 header", "mbn_stitch")
                return None
            h = MBN_V2.parse(data[: MBN_V2.sizeof()])
            parsed_size = h.data_size + MBN_V2.sizeof()
            logger.debug(
                "%s: encountered MBN v2 image, parsed_size = 0x%x",
                "mbn_stitch",
                parsed_size,
            )

        # MBN v1
        elif data_size > MBN_V1_MAGIC_SIZE and data[:MBN_V1_MAGIC_SIZE] == MBN_V1_MAGIC:
            if data_size < MBN_V1.sizeof():
                logger.error("%s: truncated MBN v1 header", "mbn_stitch")
                return None
            h = MBN_V1.parse(data[: MBN_V1.sizeof()])
            parsed_size = h.data_size + MBN_V1.sizeof()
            logger.debug(
                "%s: encountered MBN v1 image, parsed_size = 0x%x",
                "mbn_stitch",
                parsed_size,
            )

        # BIN
        elif (
            data_size > (MBN_BIN_MAGIC_SIZE + MBN_BIN_MAGIC_OFFSET)
            and data[MBN_BIN_MAGIC_OFFSET : MBN_BIN_MAGIC_OFFSET + MBN_BIN_MAGIC_SIZE] == MBN_BIN_MAGIC
        ):
            if data_size < MBN_BIN.sizeof():
                logger.error("%s: truncated MBN BIN header", "mbn_stitch")
                return None
            h = MBN_BIN.parse(data[: MBN_BIN.sizeof()])
            parsed_size = h.total_size
            logger.debug(
                "%s: encountered MBN BIN image, parsed_size = 0x%x",
                "mbn_stitch",
                parsed_size,
            )

        # ELF
        else:
            end = _elf_last_segment_end(data)
            if end is not None:
                parsed_size = end
                logger.debug(
                    "%s: encountered ELF image, parsed_size = 0x%x",
                    "mbn_stitch",
                    parsed_size,
                )
            else:
                logger.warning("Unknown file format passed to %s", "mbn_stitch")
                parsed_size = data_size

        if parsed_size != data_size:
            logger.warning(
                "%s: size mismatch for MBN data, expected 0x%x, input size 0x%x",
                "mbn_stitch",
                parsed_size,
                data_size,
            )

        stitch_offset = data_size - blob_size
        if stitch_offset < 0 or stitch_offset + blob_size > data_size:
            logger.error(
                "%s: stitch offset (0x%x) + size (0x%x) is larger than the destination (0x%x)",
                "mbn_stitch",
                stitch_offset,
                blob_size,
                data_size,
            )
            return None

        out = bytearray(data)
        logger.debug(
            "%s: stitching mbn at 0x%x, size 0x%x",
            "mbn_stitch",
            stitch_offset,
            blob_size,
        )
        out[stitch_offset : stitch_offset + blob_size] = blob
        return bytes(out)

    except ChecksumError:
        logger.exception("mbn_stitch: construct checksum error")
        return None
    except Exception:
        logger.exception("mbn_stitch")
        return None


def mbn_mav25_stitch(data: bytes, blob: bytes) -> Optional[bytes]:
    """
    Patch an ELF's last program-section area with a v7 trailer from `blob`.
    - Writes new metadata+hash at section start
    - Writes OEM sig+chain after existing dest qti sig+chain
    Returns new bytes or None.
    """
    if data is None:
        logger.error("%s: data is NULL", "mbn_mav25_stitch")
        return None
    if not data:
        logger.error("%s: data size is 0", "mbn_mav25_stitch")
        return None
    if blob is None:
        logger.error("%s: blob is NULL", "mbn_mav25_stitch")
        return None
    if not blob:
        logger.error("%s: blob size is 0", "mbn_mav25_stitch")
        return None

    data_size, blob_size = len(data), len(blob)

    kind, ehdr = _read_elf_headers(data)
    if not ehdr:
        logger.error("%s: data is not a valid ELF", "mbn_mav25_stitch")
        return None

    if blob_size < MBN_V7.sizeof():
        logger.error("%s: header is bigger than blob", "mbn_mav25_stitch")
        return None

    src = MBN_V7.parse(blob[: MBN_V7.sizeof()])
    _mbn_v7_log(src, "mbn_mav25_stitch", "src")
    if src.version != 7:
        logger.error(
            "%s: src header version (0x%x) is incorrect",
            "mbn_mav25_stitch",
            src.version,
        )
        return None
    if not _mbn_v7_header_sizes_expected(src):
        logger.warning(
            "%s: header sizes in header are unexpected (qti_metadata_size=0x%x, oem_metadata_size=0x%x, "
            "oem_signature_size=0x%x, oem_certificate_chain_size=0x%x)",
            "mbn_mav25_stitch",
            src.qti_metadata_size,
            src.oem_metadata_size,
            src.oem_signature_size,
            src.oem_certificate_chain_size,
        )

    # last PHDR (by index)
    phdrs = _read_program_headers(data, kind, ehdr)
    if not phdrs:
        return None
    last_ph = phdrs[-1]
    sect_off, sect_size = int(last_ph.p_offset), int(last_ph.p_filesz)

    if sect_off == 0:
        logger.error("%s: section has 0 offset", "mbn_mav25_stitch")
        return None
    if sect_size == 0:
        logger.error("%s: section has 0 size", "mbn_mav25_stitch")
        return None
    if sect_off + sect_size > data_size:
        logger.error(
            "%s: section (0x%x+0x%x) is bigger than the data",
            "mbn_mav25_stitch",
            sect_off,
            sect_size,
        )
        return None
    if sect_size < MBN_V7.sizeof():
        logger.error(
            "%s: dest header is bigger than the section (0x%x)",
            "mbn_mav25_stitch",
            sect_size,
        )
        return None

    dest = MBN_V7.parse(data[sect_off : sect_off + MBN_V7.sizeof()])
    _mbn_v7_log(dest, "mbn_mav25_stitch", "dest")
    if dest.version != 7:
        logger.error(
            "%s: dest header version (0x%x) is incorrect",
            "mbn_mav25_stitch",
            dest.version,
        )
        return None
    if not _mbn_v7_header_sizes_valid(dest, sect_size):
        logger.error(
            (
                "%s: sizes in dest header are invalid (common_metadata_size=0x%x, qti_metadata_size=0x%x, "
                "oem_metadata_size=0x%x, hash_table_size=0x%x, qti_signature_size=0x%x, "
                "qti_certificate_chain_size=0x%x, oem_signature_size=0x%x, "
                "oem_certificate_chain_size=0x%x)"
            ),
            "mbn_mav25_stitch",
            dest.common_metadata_size,
            dest.qti_metadata_size,
            dest.oem_metadata_size,
            dest.hash_table_size,
            dest.qti_signature_size,
            dest.qti_certificate_chain_size,
            dest.oem_signature_size,
            dest.oem_certificate_chain_size,
        )
        return None
    if not _mbn_v7_header_sizes_expected(dest):
        logger.warning(
            "%s: header sizes in dest header are unexpected (qti_metadata_size=0x%x, oem_metadata_size=0x%x, "
            "oem_signature_size=0x%x, oem_certificate_chain_size=0x%x)",
            "mbn_mav25_stitch",
            dest.qti_metadata_size,
            dest.oem_metadata_size,
            dest.oem_signature_size,
            dest.oem_certificate_chain_size,
        )

    # compute new layout from src
    new_metadata_size = MBN_V7.sizeof() + src.common_metadata_size + src.qti_metadata_size + src.oem_metadata_size
    new_metadata_and_hash_table_size = new_metadata_size + src.hash_table_size
    new_oem_sig_and_cert_chain_size = src.oem_signature_size + src.oem_certificate_chain_size

    new_oem_sig_and_cert_chain_off = (
        new_metadata_and_hash_table_size + dest.qti_signature_size + dest.qti_certificate_chain_size
    )

    # bounds
    if new_metadata_and_hash_table_size > blob_size:
        logger.error(
            "%s: new metadata (0x%x) and hash table (0x%x) are bigger than the source (0x%x)",
            "mbn_mav25_stitch",
            new_metadata_size,
            src.hash_table_size,
            blob_size,
        )
        return None
    if new_metadata_and_hash_table_size > sect_size:
        logger.error(
            "%s: new metadata (0x%x) and hash table (0x%x) are bigger than the destination (0x%x)",
            "mbn_mav25_stitch",
            new_metadata_size,
            src.hash_table_size,
            sect_size,
        )
        return None
    if new_metadata_and_hash_table_size + new_oem_sig_and_cert_chain_size > blob_size:
        logger.error(
            "%s: new OEM signature and certificate chain are bigger than the source",
            "mbn_mav25_stitch",
        )
        return None
    if new_oem_sig_and_cert_chain_off + new_oem_sig_and_cert_chain_size > sect_size:
        logger.error(
            "%s: new OEM signature and certificate chain are outside the bounds of the destination",
            "mbn_mav25_stitch",
        )
        return None

    out = bytearray(data)
    # write metadata + hash
    logger.debug(
        "%s: stitching mbn at 0x%x (0x%x bytes)",
        "mbn_mav25_stitch",
        sect_off,
        new_metadata_and_hash_table_size,
    )
    out[sect_off : sect_off + new_metadata_and_hash_table_size] = blob[:new_metadata_and_hash_table_size]

    # write OEM sig + chain
    logger.debug(
        "%s: stitching mbn at 0x%x (0x%x bytes)",
        "mbn_mav25_stitch",
        sect_off + new_oem_sig_and_cert_chain_off,
        new_oem_sig_and_cert_chain_size,
    )
    start = new_metadata_and_hash_table_size
    end = start + new_oem_sig_and_cert_chain_size

    slice_start = sect_off + new_oem_sig_and_cert_chain_off
    slice_end = sect_off + new_oem_sig_and_cert_chain_off + new_oem_sig_and_cert_chain_size
    out[slice_start:slice_end] = blob[start:end]

    return bytes(out)


# -----------------------------------------------------------------------------
# Tiny helpers (optional)
# -----------------------------------------------------------------------------


def mbn_is_valid_elf(buf: bytes) -> bool:
    return len(buf) >= EI_NIDENT and _is_valid_elf_ident(buf[:EI_NIDENT])


def mbn_is_64bit_elf(buf: bytes) -> bool:
    return len(buf) >= EI_NIDENT and buf[EI_CLASS] == ELFCLASS64



import asyncio
import contextlib
import json
import re
import sys
from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from typing import Annotated, Optional, TextIO

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import (
    ServiceProviderDep,
    async_command,
    default_json_encoder,
    print_json,
    prompt_selection,
)
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.dvt_provider import DvtProvider
from pymobiledevice3.services.dvt.instruments.sysmontap import Sysmontap

cli = InjectingTyper(
    name="process",
    help="Process monitor options.",
    no_args_is_help=True,
)

monitor_cli = InjectingTyper(
    name="monitor",
    help="Continuously stream process metrics.",
    no_args_is_help=True,
)
cli.add_typer(monitor_cli)

_BYTE_FIELDS = {
    "anonMemoryUsage",
    "diskBytesRead",
    "diskBytesWritten",
    "memAnon",
    "memCompressed",
    "memResidentSize",
    "memVirtualSize",
    "physFootprint",
    "purgeableMemory",
    "wiredMemory",
}

_PROCESS_FILTER_PATTERN = re.compile(r"(?P<key>[^=]+)=(?P<value>.*)")


class ProcessSelectionMode(str, Enum):
    PROMPT = "prompt"
    FIRST = "first"
    LAST = "last"


ProcessFilterExpressions = Annotated[
    Optional[list[str]],
    typer.Option(
        "--filter",
        "-f",
        help="Filter processes by key=value. Can be specified multiple times.",
    ),
]

ProcessKeys = Annotated[
    Optional[list[str]],
    typer.Option(
        "--key",
        "-k",
        help="Show only selected process keys for each emitted record. Can be specified multiple times.",
    ),
]

ProcessMonitorDuration = Annotated[
    Optional[int],
    typer.Option(
        "--duration",
        "-d",
        help="Maximum duration in milliseconds to run monitoring (optional)",
    ),
]

HumanReadableProcessValues = Annotated[
    bool,
    typer.Option(
        "--human",
        help="Format known byte-count fields such as physFootprint using human-readable units.",
    ),
]

ProcessResultsOutputPath = Annotated[
    Optional[Path],
    typer.Option(
        "--output",
        "-o",
        help="Output file path for JSONL format (optional, defaults to stdout)",
    ),
]


def _parse_process_filters(filter_expressions: Optional[list[str]]) -> dict[str, list[str]]:
    parsed_filters: dict[str, list[str]] = {}
    if filter_expressions:
        for raw in filter_expressions:
            match = _PROCESS_FILTER_PATTERN.fullmatch(raw)
            if match is None:
                raise typer.BadParameter(f'Invalid filter "{raw}". Expected key=value.')
            key = match.group("key")
            value = match.group("value")
            parsed_filters.setdefault(key, []).append(value)
    return parsed_filters


def _validate_process_keys(process: dict, keys: list[str]) -> None:
    unknown_keys = [key for key in keys if key not in process]
    if unknown_keys:
        raise typer.BadParameter(
            f"The process does not have the following keys: {unknown_keys}. Possible keys={list(process.keys())}"
        )


def _matches_filters(proc: dict, parsed_filters: dict[str, list[str]]) -> bool:
    if len(parsed_filters) == 0:
        return True
    return all(str(proc.get(key)) in values for key, values in parsed_filters.items())


def _select_process_output_keys(process: dict, output_keys: Optional[list[str]]) -> dict:
    if not output_keys:
        return process

    _validate_process_keys(process, output_keys)
    return {key: process[key] for key in output_keys}


def _format_byte_count(value: int) -> str:
    suffixes = ("B", "KB", "MB", "GB", "TB", "PB")
    size = float(value)
    suffix_index = 0
    while size >= 1024 and suffix_index < len(suffixes) - 1:
        size /= 1024
        suffix_index += 1
    if suffix_index == 0:
        return f"{int(size)}{suffixes[suffix_index]}"
    return f"{size:.1f}{suffixes[suffix_index]}" if size < 10 else f"{size:.0f}{suffixes[suffix_index]}"


def _humanize_process_values(process: dict) -> dict:
    humanized = dict(process)
    for key, value in humanized.items():
        if key in _BYTE_FIELDS and isinstance(value, int):
            humanized[key] = _format_byte_count(value)
    return humanized


def _serialize_process(process: dict, keys: Optional[list[str]] = None, human: bool = False) -> dict:
    selected = dict(_select_process_output_keys(process, keys))
    if human:
        selected = _humanize_process_values(selected)
    selected["timestamp"] = datetime.now(timezone.utc).isoformat()
    return selected


def _write_process(out: Optional[TextIO], process: dict) -> None:
    if out is None:
        print_json(process)
        return

    json_output = json.dumps(process, default=default_json_encoder)
    out.write(json_output + "\n")
    out.flush()


def _write_json(out: Optional[TextIO], value) -> None:
    if out is None:
        print_json(value)
        return

    out.write(json.dumps(value, sort_keys=True, indent=4, default=default_json_encoder) + "\n")
    out.flush()


def _write_status(line: str) -> None:
    print(line, flush=True)


def _describe_process(process: dict) -> str:
    name = process.get("name") or process.get("comm") or "<unknown>"
    pid = process.get("pid")
    ppid = process.get("ppid")
    return f"pid={pid}, ppid={ppid}, name={name}"


def _describe_processes(processes: list[dict]) -> str:
    return "; ".join(_describe_process(process) for process in processes)


def _duration_elapsed(start_time: float, duration_ms: Optional[int]) -> bool:
    if duration_ms is None:
        return False
    return ((asyncio.get_running_loop().time() - start_time) * 1000) >= duration_ms


def _should_skip_first_snapshot(keys: Optional[list[str]]) -> bool:
    # The first sample does not contain initialized cpuUsage values
    return (keys is None) or ("cpuUsage" in keys)


async def iter_processes(sysmon: Sysmontap, skip_first_snapshot: bool = False):
    should_skip_snapshot = skip_first_snapshot
    async for process_snapshot in sysmon.iter_processes():
        if should_skip_snapshot:
            should_skip_snapshot = False
            continue
        yield process_snapshot


def _process_sort_key(process: dict) -> tuple:
    # Sort from oldest to newest process. "first" picks the oldest match and "last" picks the newest match.
    start_abs_time = process.get("startAbsTime")
    if not isinstance(start_abs_time, int):
        start_abs_time = -1
    pid = process.get("pid")
    if not isinstance(pid, int):
        pid = -1
    name = process.get("name") or process.get("comm") or ""
    return start_abs_time, pid, str(name)


def _select_process_from_candidates(processes: list[dict], selection_mode: ProcessSelectionMode) -> dict:
    if len(processes) == 1:
        return processes[0]

    sorted_processes = sorted(processes, key=_process_sort_key)

    if selection_mode == ProcessSelectionMode.FIRST:
        return sorted_processes[0]
    if selection_mode == ProcessSelectionMode.LAST:
        return sorted_processes[-1]

    if not sys.stdin.isatty():
        raise typer.BadParameter(
            f'Multiple processes matched the given filters. Re-run with "--choose first", "--choose last", or refine the filters. '
            f"Matches: {_describe_processes(processes)}"
        )

    selection_index = prompt_selection(
        [_describe_process(process) for process in sorted_processes],
        "Choose process to monitor",
        idx=True,
    )
    return sorted_processes[selection_index]


def _get_process_identifier(process: dict) -> tuple[str, object]:
    if process.get("uniqueID") is not None:
        return "uniqueID", process["uniqueID"]
    if process.get("startAbsTime") is not None:
        return "startAbsTime", process["startAbsTime"]
    return "pid", process["pid"]


def _matches_selected_process(process: dict, selected_process_identifier: tuple[str, object]) -> bool:
    identifier_key, identifier_value = selected_process_identifier
    return process.get(identifier_key) == identifier_value


def _select_process_from_snapshot(
    process_snapshot: list[dict],
    parsed_filters: dict[str, list[str]],
    selection_mode: ProcessSelectionMode,
) -> dict:
    if process_snapshot:
        # All process entries in a sysmon sample share the same schema, so validating one entry is sufficient.
        _validate_process_keys(process_snapshot[0], list(parsed_filters))

    matching_processes = [process for process in process_snapshot if _matches_filters(process, parsed_filters)]
    if len(matching_processes) == 0:
        raise typer.BadParameter("Failed to find a process matching the given filters in the current snapshot")

    return _select_process_from_candidates(matching_processes, selection_mode)


async def _select_process_from_sysmon(
    dvt, parsed_filters: dict[str, list[str]], keys: Optional[list[str]], selection_mode: ProcessSelectionMode
) -> dict:
    async with await Sysmontap.create(dvt) as selection_sysmon:
        async for process_snapshot in iter_processes(
            selection_sysmon, skip_first_snapshot=_should_skip_first_snapshot(keys)
        ):
            # All process entries in a sysmon sample share the same schema, so validating one entry is sufficient.
            _validate_process_keys(process_snapshot[0], keys or [])
            return _select_process_from_snapshot(process_snapshot, parsed_filters, selection_mode)

    raise typer.BadParameter("Failed to collect a process snapshot")


@cli.command("single")
@async_command
async def sysmon_process_single(
    service_provider: ServiceProviderDep,
    filter_expressions: ProcessFilterExpressions = None,
    keys: ProcessKeys = None,
    human: HumanReadableProcessValues = False,
    output: ProcessResultsOutputPath = None,
) -> None:
    """show a single snapshot of currently running processes."""
    with contextlib.ExitStack() as stack:
        out = stack.enter_context(open(output, "w")) if output else None
        await sysmon_process_single_task(service_provider, filter_expressions, keys, human, out)


async def sysmon_process_single_task(
    service_provider: ServiceProviderDep,
    filter_expressions: Optional[list[str]] = None,
    keys: Optional[list[str]] = None,
    human: bool = False,
    out: Optional[TextIO] = None,
) -> None:
    parsed_filters = _parse_process_filters(filter_expressions)
    result = []
    async with (
        DvtProvider(service_provider) as dvt,
        DeviceInfo(dvt) as device_info,
        await Sysmontap.create(dvt) as sysmon,
    ):
        async for process_snapshot in iter_processes(sysmon, skip_first_snapshot=_should_skip_first_snapshot(keys)):
            if process_snapshot and parsed_filters:
                # All process entries in a sysmon sample share the same schema, so validating one entry is sufficient.
                _validate_process_keys(process_snapshot[0], list(parsed_filters))

            for process in process_snapshot:
                if not _matches_filters(process, parsed_filters):
                    continue

                process = dict(process)
                process["execName"] = await device_info.execname_for_pid(process["pid"])
                result.append(_serialize_process(process, keys, human))

            break

    if len(result) == 0:
        if parsed_filters:
            raise typer.BadParameter("Failed to find a process matching the given filters in the current snapshot")
        raise typer.BadParameter("Failed to find any processes in the current snapshot")

    _write_json(out, result)


@monitor_cli.command("threshold")
@async_command
async def sysmon_process_monitor_threshold(
    service_provider: ServiceProviderDep,
    threshold: Annotated[float, typer.Argument(help="Minimum cpuUsage value to emit")],
    keys: ProcessKeys = None,
    duration: ProcessMonitorDuration = None,
    human: HumanReadableProcessValues = False,
    output: ProcessResultsOutputPath = None,
) -> None:
    """Continuously monitor processes above a cpuUsage threshold."""

    with contextlib.ExitStack() as stack:
        out = stack.enter_context(open(output, "w")) if output else None
        await sysmon_process_monitor_threshold_task(service_provider, threshold, duration, keys, human, out)


@monitor_cli.command("process")
@async_command
async def sysmon_process_monitor_process(
    service_provider: ServiceProviderDep,
    filter_expressions: ProcessFilterExpressions = None,
    keys: ProcessKeys = None,
    choose: Annotated[
        ProcessSelectionMode,
        typer.Option(
            "--choose",
            help=(
                'How to resolve multiple matching processes: "prompt" asks interactively; '
                '"first" selects the oldest matching process; "last" selects the newest matching process. '
                "Automatic ordering is by startAbsTime, then pid, then name."
            ),
        ),
    ] = ProcessSelectionMode.PROMPT,
    interval: Annotated[
        int,
        typer.Option(
            "--interval",
            "-i",
            help="Minimum interval in milliseconds between outputs (optional)",
        ),
    ] = Sysmontap.DEFAULT_INTERVAL_MS,
    duration: ProcessMonitorDuration = None,
    human: HumanReadableProcessValues = False,
    output: ProcessResultsOutputPath = None,
) -> None:
    """Continuously monitor one process selected from the current snapshot by key=value filters."""

    with contextlib.ExitStack() as stack:
        out = stack.enter_context(open(output, "w")) if output else None
        await sysmon_process_monitor_process_task(
            service_provider, filter_expressions, interval, duration, choose, keys, human, out
        )


async def sysmon_process_monitor_threshold_task(
    service_provider: ServiceProviderDep,
    threshold: float,
    duration: Optional[int] = None,
    keys: Optional[list[str]] = None,
    human: bool = False,
    out: Optional[TextIO] = None,
) -> None:
    """Continuously monitor processes above a cpuUsage threshold."""

    start_time = None

    async with DvtProvider(service_provider) as dvt, await Sysmontap.create(dvt) as sysmon:
        async for process_snapshot in iter_processes(sysmon, skip_first_snapshot=True):
            if start_time is None:
                start_time = asyncio.get_running_loop().time()

            if _duration_elapsed(start_time, duration):
                break

            if process_snapshot:
                # All process entries in a sysmon sample share the same schema, so validating one entry is sufficient.
                _validate_process_keys(process_snapshot[0], keys or [])

            for process in process_snapshot:
                if process.get("cpuUsage") is None or process["cpuUsage"] < threshold:
                    continue

                _write_process(out, _serialize_process(process, keys, human))

            if _duration_elapsed(start_time, duration):
                # Avoid waiting for the next snapshot
                break


async def sysmon_process_monitor_process_task(
    service_provider: ServiceProviderDep,
    filter_expressions: Optional[list[str]] = None,
    interval: int = Sysmontap.DEFAULT_INTERVAL_MS,
    duration: Optional[int] = None,
    choose: ProcessSelectionMode = ProcessSelectionMode.PROMPT,
    keys: Optional[list[str]] = None,
    human: bool = False,
    out: Optional[TextIO] = None,
) -> None:
    """Continuously monitor one process selected from the current snapshot by key=value filters."""

    parsed_filters = _parse_process_filters(filter_expressions)

    async with DvtProvider(service_provider) as dvt:
        selected_process = await _select_process_from_sysmon(dvt, parsed_filters, keys, choose)
        selected_process_identifier = _get_process_identifier(selected_process)
        selected_process_description = _describe_process(selected_process)

        _write_status(f"Monitoring {selected_process_description}")

        monitoring_start_time = None
        async with await Sysmontap.create(dvt, interval=interval) as monitor_sysmon:
            async for process_snapshot in iter_processes(
                monitor_sysmon, skip_first_snapshot=_should_skip_first_snapshot(keys)
            ):
                if monitoring_start_time is None:
                    monitoring_start_time = asyncio.get_running_loop().time()

                if _duration_elapsed(monitoring_start_time, duration):
                    break

                if len(process_snapshot) == 0:
                    continue

                # All process entries in a sysmon sample share the same schema, so validating one entry is sufficient.
                _validate_process_keys(process_snapshot[0], keys or [])

                selected_process_matches = [
                    process
                    for process in process_snapshot
                    if _matches_selected_process(process, selected_process_identifier)
                ]

                if len(selected_process_matches) == 0:
                    _write_status(f"Selected process exited: {selected_process_description}")
                    break

                if len(selected_process_matches) > 1:
                    matching_processes_description = [
                        _describe_process(process) for process in selected_process_matches
                    ]
                    raise typer.BadParameter(
                        f"Selected process identity is ambiguous for {selected_process_description!r}. "
                        f"Matching processes: {matching_processes_description}. Please refine the filters."
                    )

                _write_process(out, _serialize_process(selected_process_matches[0], keys, human))

                if _duration_elapsed(monitoring_start_time, duration):
                    # Avoid waiting for the next snapshot
                    break



import dataclasses
import uuid
from pathlib import Path

from tqdm import tqdm

from pymobiledevice3.remote.remote_service import RemoteService
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService


@dataclasses.dataclass
class DSCFile:
    file_path: str
    file_size: int


class RemoteFetchSymbolsService(RemoteService):
    SERVICE_NAME = "com.apple.dt.remoteFetchSymbols"

    def __init__(self, rsd: RemoteServiceDiscoveryService):
        super().__init__(rsd, self.SERVICE_NAME)

    async def get_dsc_file_list(self) -> list[DSCFile]:
        files: list[DSCFile] = []
        response = await self.service.send_receive_request({
            "XPCDictionary_sideChannel": uuid.uuid4(),
            "DSCFilePaths": [],
        })
        file_count = response["DSCFilePaths"]
        for _i in range(file_count):
            response = await self.service.receive_response()
            response = response["DSCFilePaths"]
            file_transfer = response["fileTransfer"]
            expected_length = file_transfer["expectedLength"]
            file_path = response["filePath"]
            files.append(DSCFile(file_path=file_path, file_size=expected_length))
        return files

    async def download(self, out: Path) -> None:
        files = await self.get_dsc_file_list()
        with tqdm(total=len(files), dynamic_ncols=True, desc="Downloading DSC") as total_pb:
            for i, file in enumerate(files):
                total_pb.set_description("Downloading DSC")
                out_file = out / file.file_path[1:]  # trim the "/" prefix
                out_file.parent.mkdir(parents=True, exist_ok=True)
                with (
                    open(out_file, "wb") as f,
                    tqdm(total=files[i].file_size, dynamic_ncols=True, desc=file.file_path.rsplit("/", 1)[-1]) as pb,
                ):
                    async for chunk in self.service.iter_file_chunks(files[i].file_size, file_idx=i):
                        f.write(chunk)
                        pb.update(len(chunk))
                total_pb.update(1)



from pathlib import Path
from typing import Annotated

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command
from pymobiledevice3.services.afc import AfcService, AfcShell

cli = InjectingTyper(
    name="afc",
    help="Browse, push, and pull files via the AFC service (/var/mobile/Media).",
    no_args_is_help=True,
)


@cli.command("shell")
def afc_shell(service_provider: ServiceProviderDep) -> None:
    """Open an interactive AFC shell rooted at /var/mobile/Media."""
    AfcShell.create(service_provider)


@cli.command("pull")
@async_command
async def afc_pull(
    service_provider: ServiceProviderDep,
    remote_file: str,
    local_file: Path,
    ignore_errors: Annotated[
        bool,
        typer.Option(
            "--ignore-errors",
            "-i",
            help="Continue downloading even if some files error (best-effort pull).",
        ),
    ],
) -> None:
    """Download a remote path under /var/mobile/Media to the local filesystem."""
    async with AfcService(lockdown=service_provider) as afc:
        await afc.pull(remote_file, str(local_file), ignore_errors=ignore_errors)


@cli.command("push")
@async_command
async def afc_push(service_provider: ServiceProviderDep, local_file: Path, remote_file: str) -> None:
    """Upload a local file into /var/mobile/Media."""
    async with AfcService(lockdown=service_provider) as afc:
        await afc.push(str(local_file), remote_file)


@cli.command("ls")
@async_command
async def afc_ls(
    service_provider: ServiceProviderDep,
    remote_file: str,
    recursive: Annotated[
        bool,
        typer.Option("--recursive", "-r", help="Recurse into subdirectories when listing."),
    ] = False,
) -> None:
    """List files under /var/mobile/Media (optionally recursively)."""
    async with AfcService(lockdown=service_provider) as afc:
        async for path in afc.dirlist(remote_file, -1 if recursive else 1):
            print(path)


@cli.command("rm")
@async_command
async def afc_rm(service_provider: ServiceProviderDep, remote_file: str) -> None:
    """Delete a file under /var/mobile/Media."""
    async with AfcService(lockdown=service_provider) as afc:
        await afc.rm(remote_file)



from __future__ import annotations

import pytest

from pymobiledevice3.dtx._sender import _DTXSenderMixin
from pymobiledevice3.dtx.message import DTXMessage, DTXMessageType
from pymobiledevice3.dtx.ns_types import NSError


class _DummySender(_DTXSenderMixin):
    def __init__(self) -> None:
        self.sent: list[DTXMessage] = []

    async def _send_message(self, message: DTXMessage) -> None:
        self.sent.append(message)


@pytest.mark.asyncio
async def test_send_reply_ack_has_no_payload_or_aux() -> None:
    sender = _DummySender()

    await sender.send_reply_ack(channel_code=7, msg_id=11, conv_idx=2)

    assert len(sender.sent) == 1
    msg = sender.sent[0]
    assert msg.type == DTXMessageType.OK
    assert msg.payload is None
    assert list(msg.aux) == []
    assert len(msg.payload_data) == 0
    assert len(msg.aux_data) == 0


@pytest.mark.asyncio
async def test_send_reply_defaults_none_aux_to_empty_aux() -> None:
    sender = _DummySender()

    await sender.send_reply(channel_code=3, msg_id=12, conv_idx=2, payload={"key": "value"})

    assert len(sender.sent) == 1
    msg = sender.sent[0]
    assert msg.type == DTXMessageType.OBJECT
    assert msg.payload == {"key": "value"}
    assert list(msg.aux) == []
    assert len(msg.aux_data) == 0


@pytest.mark.asyncio
async def test_send_reply_error_rejects_non_empty_aux() -> None:
    sender = _DummySender()
    error = NSError(1, "DTXMessage", {"NSLocalizedDescription": "failure"})

    with pytest.raises(AssertionError, match="ERROR replies must not have aux arguments"):
        await sender._send_reply(5, 10, 2, DTXMessageType.ERROR, payload=error, aux_args=[1])



import io
import logging
from typing import Any, Union

from bpylist2 import archiver
from construct import Bytes, ConstructError, Peek

from .primitives import PNULL, PBuf, PDict, _primitive_value_con, _PrimitiveBase

logger = logging.getLogger(__name__)
_test_eof = Peek(Bytes(1))


class MessageAux(list[Any]):
    """An adapter that parse DISPTACH arguments from/to primitive dictionaries"""

    @classmethod
    def parse(cls, obj: Union[bytes, bytearray, memoryview], context, path):
        if len(obj) == 0:
            # interpret empty buffers as an empty list
            return []
        stream = io.BytesIO(obj)
        primitive_dict = _primitive_value_con._parse(stream, context, f"{path}.aux")
        if not isinstance(primitive_dict, dict):
            raise ConstructError(
                f"Expected a dictionary for MessageAux, got {type(primitive_dict).__name__}", path=path
            )
        if len(primitive_dict) != 1 or not isinstance(primitive_dict.get(PNULL), list):
            raise ConstructError(
                f"Expected a dictionary with a single PNULL key mapping to a list, got {primitive_dict}", path=path
            )
        converted_list = []
        for arg in primitive_dict[PNULL]:
            assert isinstance(arg, _PrimitiveBase), (
                f"Expected all arguments to be primitive types, got {type(arg).__name__}"
            )
            if arg is PNULL:
                logger.warning(f"Received a PNULL argument instead of an empty buffer, dict={primitive_dict!r}")
                converted_list.append(None)
            elif not isinstance(arg, PBuf):
                # primitive types are passed as-is
                # they extens the built-in types, so we can just pass them through without converting
                converted_list.append(arg)
            else:
                if len(arg) == 0:
                    converted_list.append(None)  # interpret empty buffers as None
                else:
                    try:
                        converted_list.append(archiver.unarchive(arg))
                    except Exception as e:
                        raise ConstructError(
                            f"Failed to unarchive argument from buffer: {e}, buf={arg!r}", path=path
                        ) from e

        return converted_list

    @classmethod
    def build(cls, obj, context, path) -> bytes:
        if not obj:
            # nothing is written ¯\_(ツ)_/¯
            return b""
        converted_list = []
        # translate everything that is not a primitive into a NSKeyedArchiver-encoded blob
        for arg in obj:
            if isinstance(arg, _PrimitiveBase):
                converted_list.append(arg)
            elif arg is None:
                # An empty buffer is placed instead of a PrimitiveNull ¯\_(ツ)_/¯
                converted_list.append(PBuf(b""))
            else:
                try:
                    converted_list.append(PBuf(archiver.archive(arg)))
                except Exception as e:
                    raise ConstructError(
                        f"Failed to archive argument {arg} of type <{type(arg).__name__}>: {e}", path=path
                    ) from e

        pdict = PDict({PNULL: converted_list})
        stream = io.BytesIO()
        _primitive_value_con._build(pdict, stream, context, f"{path}.aux")
        return stream.getvalue()
